"""(De)serializers for primitive types."""

from umbi.datatypes import Primitive, PrimitiveType
from bitstring import BitArray


# Booleans


def bytes_to_bool(data: bytes) -> bool:
    # should not be needed in umbi
    raise NotImplementedError("unexpected need to decode a boolean from bytes")


def bool_to_bytes(value: bool) -> bytes:
    # should not be needed in umbi
    raise NotImplementedError("unexpected need to encode a boolean into bytes")


def bits_to_bool(bits: BitArray) -> bool:
    """Convert a bitstring to a boolean."""
    return bits.uint != 0


def bool_to_bits(value: bool, num_bits: int = 1) -> BitArray:
    """Convert a boolean to a bitstring."""
    value_uint = 1 if value else 0
    return BitArray(uint=value_uint, length=num_bits)


# Strings


def bytes_to_string(bytestring: bytes) -> str:
    """Convert a bytestring to a utf-8 string."""
    return bytestring.decode("utf-8")


def string_to_bytes(value: str) -> bytes:
    """Convert a utf-8 string to a bytestring."""
    return value.encode("utf-8")


def string_to_bits(value: str) -> BitArray:
    # should not be needed in umbi
    raise NotImplementedError("unexpected need to encode a string into a bitstring")


def bits_to_string(bits: BitArray) -> str:
    # should not be needed in umbi
    raise NotImplementedError("unexpected need to decode a string from a bitstring")


# API


def bytes_to_primitive(data: bytes, value_type: PrimitiveType) -> Primitive:
    return {
        PrimitiveType.BOOL: bytes_to_bool,
        PrimitiveType.STRING: bytes_to_string,
    }[value_type](data)


def primitive_to_bytes(value: Primitive, value_type: PrimitiveType) -> bytes:
    return {
        PrimitiveType.BOOL: bool_to_bytes,
        PrimitiveType.STRING: string_to_bytes,
    }[value_type](value)


def bits_to_primitive(bits: BitArray, value_type: PrimitiveType) -> Primitive:
    return {
        PrimitiveType.BOOL: bits_to_bool,
        PrimitiveType.STRING: bits_to_string,
    }[value_type](bits)


def primitive_to_bits(value: Primitive, value_type: PrimitiveType, num_bits: int) -> BitArray:
    return {
        PrimitiveType.BOOL: lambda v: bool_to_bits(v, num_bits),
        PrimitiveType.STRING: string_to_bits,
    }[value_type](value)


__all__ = [
    "bytes_to_primitive",
    "primitive_to_bytes",
    "bits_to_primitive",
    "primitive_to_bits",
]



"""ATS model examples."""

from umbi.examples.ats.grid import grid_ats_from_string
from umbi.examples.ats.random_game import random_game_ats
from umbi.examples.ats.random_walk import random_walk_ats

__all__ = [
    "grid_ats_from_string",
    "random_game_ats",
    "random_walk_ats",
]



"""
Numeric datatypes (integers, floats, rationals) that allow promotions.
"""

import enum
from fractions import Fraction

""" Numeric primitive types. """


class NumericPrimitiveType(str, enum.Enum):
    INT = "int"
    UINT = "uint"
    DOUBLE = "double"
    RATIONAL = "rational"


""" Alias for primitive numeric objects. """
NumericPrimitive = int | float | Fraction


def is_integer_type(type: NumericPrimitiveType) -> bool:
    return type in [NumericPrimitiveType.INT, NumericPrimitiveType.UINT]


def assert_integer_type(type: NumericPrimitiveType):
    assert is_integer_type(type), f"not an integer type: {type}"


def integer_type_signed(type: NumericPrimitiveType) -> bool:
    assert_integer_type(type)
    return type == NumericPrimitiveType.INT


def numeric_primitive_type_of(value: NumericPrimitive) -> NumericPrimitiveType:
    """Determine the numeric type of a given value."""
    if isinstance(value, int):
        return NumericPrimitiveType.INT
    elif isinstance(value, float):
        return NumericPrimitiveType.DOUBLE
    else:  # isinstance(value, Fraction):
        return NumericPrimitiveType.RATIONAL


def numeric_primitive_promotion_type(types: set[NumericPrimitiveType]) -> NumericPrimitiveType:
    """Determine the common numeric type from a set of numeric types. Used for type promotion."""
    assert len(types) > 0, "cannot determine common numeric type of empty set"
    assert all(isinstance(t, NumericPrimitiveType) for t in types), f"non-numeric types found in set: {types}"
    if NumericPrimitiveType.RATIONAL in types:
        return NumericPrimitiveType.RATIONAL
    elif NumericPrimitiveType.DOUBLE in types:
        return NumericPrimitiveType.DOUBLE
    elif NumericPrimitiveType.INT in types:
        return NumericPrimitiveType.INT
    else:  # only NumericPrimitiveType.UINT in types
        return NumericPrimitiveType.UINT


def promote_numeric_primitive_to(value: NumericPrimitive, target_type: NumericPrimitiveType) -> NumericPrimitive:
    """
    Promote a primitive numeric value to the target type.
    Promotion rules: int -> double -> rational
    """
    if numeric_primitive_type_of(value) == target_type:
        return value
    if target_type == NumericPrimitiveType.INT:
        raise ValueError(f"cannot promote value {value} to int")
    elif target_type == NumericPrimitiveType.DOUBLE:
        assert isinstance(value, int), f"cannot promote value {value} to double"
        return float(value)
    else:
        assert target_type == NumericPrimitiveType.RATIONAL, f"unexpected target type: {target_type}"
        if isinstance(value, int):
            return Fraction(value, 1)
        elif isinstance(value, float):
            return Fraction.from_float(value)
        else:
            raise ValueError(f"cannot promote value {value} to rational")



"""
Type schema and result classes for UMBI types.
"""

from typing import Any
from marshmallow import fields, post_load
from .json_schema import JsonSchema
import umbi.datatypes
from umbi.binary.sized_type import SizedType


class FieldType(fields.String):
    def _deserialize(self, value: Any, attr: Any, data: Any, **kwargs: Any) -> Any:
        # Validate the input value directly without using parent's validator
        # which expects a String return type
        if not isinstance(value, str):
            raise ValueError(f"Expected string value, got {type(value)}")

        string_value = value

        # Validate against allowed types
        allowed_types = (
            [t.value for t in umbi.datatypes.PrimitiveType]
            + [t.value for t in umbi.datatypes.NumericPrimitiveType]
            + [f"{t.value}-interval" for t in umbi.datatypes.NumericPrimitiveType]
        )
        if string_value not in allowed_types:
            raise ValueError(f"Must be one of: {', '.join(allowed_types)}. Got: {string_value}")

        # Convert string to ScalarType enum or IntervalType
        for enum_class in [
            umbi.datatypes.PrimitiveType,
            umbi.datatypes.NumericPrimitiveType,
        ]:
            try:
                return enum_class(string_value)
            except ValueError:
                pass
        try:
            return umbi.datatypes.IntervalType.from_string(string_value)
        except ValueError:
            raise ValueError(f"Unknown type: {string_value}")

    def _serialize(self, value: Any, attr: Any, obj: Any, **kwargs: Any) -> str:
        assert isinstance(value, umbi.datatypes.ScalarType), "expected a ScalarType value"
        if isinstance(value, (umbi.datatypes.PrimitiveType, umbi.datatypes.NumericPrimitiveType)):
            return value.value
        else:  # isinstance(value, umbi.datatypes.IntervalType)
            return str(value)


class SizedTypeSchema(JsonSchema):
    type = FieldType(data_key="type", required=True)
    size = fields.Integer(data_key="size", required=False)

    @post_load
    def make_object(self, data: dict, **kwargs: Any) -> SizedType:
        obj = super().make_object(data, **kwargs)
        assert isinstance(obj.type, umbi.datatypes.ScalarType), f"expected ScalarType, got {type(obj.type)}"
        return SizedType(type=obj.type, size_bits=obj.size)

    def dump(self, obj: Any, *args: Any, **kwargs: Any) -> Any:
        assert isinstance(obj, SizedType)
        obj_dict = {
            "type": FieldType()._serialize(obj.type, "type", obj),
            "size": obj.size_bits,
        }
        return obj_dict



"""Composite datatype definitions."""

from umbi.datatypes import IntervalType, PrimitiveType, Scalar

from .sized_type import SizedType, UINT64
from .scalar import bits_to_scalar, scalar_to_bits
from .utils import split_bytes

from bitstring import BitArray
from dataclasses import dataclass, field as dataclass_field


# Struct type definitions

# alias for instances of StructType, as a mapping from attribute names to (optional) values
Struct = dict[str, Scalar | None]


@dataclass(frozen=True)
class StructPadding:
    """Padding bits in a composite datatype."""

    padding: int

    def validate(self):
        if self.padding <= 0:
            raise ValueError(f"Padding must be positive ({self.padding})")

    @property
    def size_bits(self) -> int:
        return self.padding


@dataclass
class StructAttribute:
    """A variable field in a composite datatype."""

    name: str
    sized_type: SizedType
    is_optional: bool = (
        False  # if True, then the value is preceded with a presence flag, 1 = present, 0 = present but must be ignored
    )
    lower: int | None = None  # lower bound (for numeric types)
    upper: int | None = None  # upper bound (for numeric types)
    offset: int | None = None  # lower value offset (for numeric types)

    @property
    def size_bits(self) -> int:
        size_bits = self.sized_type.size_bits
        if self.is_optional:
            size_bits += 1  # presence flag
        return size_bits

    def validate(self):
        self.sized_type.validate()
        if isinstance(self.sized_type.type, IntervalType):
            raise ValueError(f"Struct attribute cannot be of interval type: {self.sized_type.type}")
        if self.offset is not None and self.offset != 0:
            raise NotImplementedError("non-zero offset validation not implemented")
        # TODO validate bounds for numeric types


""" Alias for either struct field type """
StructField = StructPadding | StructAttribute


@dataclass
class StructType:
    """A composite datatype consisting of attributes and paddings."""

    fields: list[StructField] = dataclass_field(default_factory=list)

    def validate(self):
        for item in self.fields:
            item.validate()

    @property
    def attributes(self):
        """Iterate over attributes only, skipping padding fields."""
        return [f for f in self.fields if isinstance(f, StructAttribute)]

    @property
    def size_bits(self) -> int:
        return sum(f.size_bits for f in self.fields)

    @property
    def is_byte_aligned(self) -> bool:
        return self.size_bits % 8 == 0

    @property
    def size_bytes(self) -> int:
        return (self.size_bits + 7) // 8  # ceiling division

    @property
    def contains_strings(self) -> bool:
        return any(isinstance(f, StructAttribute) and f.sized_type.type == PrimitiveType.STRING for f in self.fields)

    def __str__(self) -> str:
        lines = ["struct:"]
        for f in self.fields:
            lines.append(f"  {f}")
        return "\n".join(lines)

    def __iter__(self):
        return iter(self.fields)

    @property
    def num_bits_to_pad(self) -> int:
        """Calculate the number of padding bits needed to align current struct to a full byte."""
        return (8 - self.size_bits % 8) % 8

    def add_padding(self, num_bits: int) -> None:
        """Add padding bits to the struct."""
        assert num_bits > 0
        self.fields.append(StructPadding(padding=num_bits))

    def pad_to_byte(self) -> None:
        """Add padding bits to align struct to a full byte."""
        if self.is_byte_aligned:
            return
        self.add_padding(self.num_bits_to_pad)

    def add_attribute(self, name: str, sized_type: SizedType) -> None:
        """Add an attribute field to the struct."""
        self.fields.append(StructAttribute(name=name, sized_type=sized_type))


# Struct (de)serialization


class StructUnpacker:
    """Utility class for unpacking composite datatypes from a bytestring."""

    def __init__(self, bytestring: bytes) -> None:
        self.bytestring = bytestring  # input bytestring, little-endian order
        self.buffer = BitArray()  # bit buffer, MSB at [0]

    def assert_buffer_empty(self):
        if len(self.buffer) > 0:
            raise RuntimeError("expected the buffer to be empty")

    def align_buffer(self, num_bits: int):
        """Ensure the buffer has the required number of bits."""
        if len(self.buffer) >= num_bits:
            return
        num_bytes_needed = (num_bits - len(self.buffer) + 7) // 8
        assert len(self.bytestring) >= num_bytes_needed, "not enough data to fill the buffer"
        for _ in range(num_bytes_needed):  # read one byte at a time since the bytestring is little-endian
            next_bytes, self.bytestring = split_bytes(self.bytestring, 1)
            self.buffer = BitArray(bytes=next_bytes) + self.buffer  # prepend to the buffer to keep MSB at [0]

    def extract_from_buffer(self, num_bits: int):
        """Extract the given number of bits from the buffer."""
        assert num_bits >= 0
        self.align_buffer(num_bits)
        bits = self.buffer[-num_bits:]  # remove from the end (LSB side)
        self.buffer = self.buffer[:-num_bits]
        return bits

    def skip_padding(self, field: StructPadding):
        self.extract_from_buffer(field.padding)

    def unpack_attribute(self, field: StructAttribute) -> Scalar | None:
        """
        Unpack a single field from the buffer.
        :return: the unpacked value, or None if the field is optional and not present
        """
        sized_type = field.sized_type
        if field.sized_type.type == PrimitiveType.STRING:
            # a string is stored as an index to the list of strings
            sized_type = UINT64
        bits = self.extract_from_buffer(sized_type.size_bits)
        if field.is_optional:
            is_present_bit = bits[-1]  # remove from the end (LSB side)
            bits = bits[:-1]
            if not is_present_bit:
                return None
        return bits_to_scalar(bits, sized_type.type)

    def unpack_struct(self, value_type: StructType) -> Struct:
        name_value: Struct = {}
        for field in value_type:
            if isinstance(field, StructPadding):
                self.skip_padding(field)
                continue
            name_value[field.name] = self.unpack_attribute(field)
        self.assert_buffer_empty()
        return name_value


class StructPacker:
    """Utility class for packing structs into a bytestring."""

    def __init__(self) -> None:
        self.buffer = BitArray()  # bit buffer, MSB at [0]
        self.bytestring = bytes()  # output bytestring, little-endian order

    def assert_buffer_empty(self):
        if len(self.buffer) > 0:
            raise RuntimeError("expected the buffer to be empty")

    def flush_buffer(self):
        """Flush full bytes from the buffer to the bytestring."""
        while len(self.buffer) >= 8:  # append one byte at a time so that the bytestring is little-endian
            self.buffer, bits = (
                self.buffer[:-8],
                self.buffer[-8:],
            )  # get new bits from the end (LSB side)
            new_bytes = bits.tobytes()
            self.bytestring += new_bytes

    def append_to_buffer(self, bits: BitArray):
        """Append bits to the buffer flush full bytes to the bytestring."""
        self.buffer = bits + self.buffer  # prepend to the start (MSB side)
        self.flush_buffer()

    def add_padding(self, field: StructPadding):
        """Add padding bits to the buffer."""
        self.append_to_buffer(BitArray(uint=0, length=field.padding))

    def pack_attribute(self, field: StructAttribute, value: Scalar | None):
        """Pack a single attribute into the buffer or the bytestring."""
        sized_type = field.sized_type
        if field.sized_type.type == PrimitiveType.STRING:
            # a string is stored as an index to the list of strings
            sized_type = UINT64
        value_present = value is not None
        if not value_present:
            assert field.is_optional, "only optional fields can have None value"
            value = 0  # use an arbitrary value
        bits = scalar_to_bits(value, sized_type.type, sized_type.size_bits)
        if field.is_optional:
            presence_value = 1 if value_present else 0
            presence_bit = BitArray(uint=presence_value, length=1)  # set presence bit
            bits = bits + presence_bit  # append presence bit to the end (LSB side)
        self.append_to_buffer(bits)

    def pack_struct(self, value_sized_type: StructType, struct: Struct) -> bytes:
        assert value_sized_type.is_byte_aligned, "cannot pack a struct that is not byte-aligned"
        for field in value_sized_type:
            if isinstance(field, StructPadding):
                self.add_padding(field)
            else:
                assert isinstance(field, StructAttribute)
                self.pack_attribute(field, struct[field.name])
        self.assert_buffer_empty()
        return self.bytestring


# API


def bytes_to_struct(bytestring: bytes, value_type: StructType) -> Struct:
    """Unpack a BitArray to a composite datatype."""
    return StructUnpacker(bytestring).unpack_struct(value_type)


def struct_to_bytes(struct: Struct, value_type: StructType) -> bytes:
    """Convert a composite datatype to a BitArray."""
    return StructPacker().pack_struct(value_type, struct)



#!/usr/bin/env python3
"""
umbi demo: A random walk ATS.
"""

import argparse
import logging
import pathlib
from fractions import Fraction

import umbi

logger = logging.getLogger(__name__)


def random_walk_ats(num_states: int) -> umbi.ats.ExplicitAts:
    # create ATS
    ats = umbi.ats.ExplicitAts()

    # basic parameters
    ats.time = umbi.ats.TimeType.DISCRETE
    ats.num_players = 0
    ats.num_states = num_states
    ats.num_initial_states = 1
    ats.set_initial_states([num_states // 2])  # start in the middle state

    # two actions: left (0) and right (1), each with 0.9 success prob
    ats.num_choice_actions = 2
    ats.choice_action_to_name = ["left", "right"]
    ats.num_choices = 2 * ats.num_states  # each state has 2 choices (left and right)
    ats.num_branches = 2 * ats.num_choices  # each choice has 2 branches (succeed or fail)

    # build structure
    ats.state_to_choice = []
    ats.choice_to_choice_action = []
    ats.choice_to_branches = []
    ats.branch_to_target = []
    ats.branch_to_probability = []

    for state in range(ats.num_states):
        ats.state_to_choice.append(len(ats.choice_to_choice_action))

        # left action
        ats.choice_to_choice_action.append(0)
        ats.choice_to_branches.append(len(ats.branch_to_target))
        left = max(0, state - 1)
        ats.branch_to_target.extend([left, state])
        ats.branch_to_probability.extend([Fraction(9, 10), Fraction(1, 10)])

        # right action
        ats.choice_to_choice_action.append(1)
        ats.choice_to_branches.append(len(ats.branch_to_target))
        right = min(ats.num_states - 1, state + 1)
        ats.branch_to_target.extend([right, state])
        ats.branch_to_probability.extend([Fraction(9, 10), Fraction(1, 10)])

    ats.state_to_choice.append(len(ats.choice_to_choice_action))
    ats.choice_to_branches.append(len(ats.branch_to_target))

    ats.state_is_markovian = [True] * ats.num_states
    ats.state_to_exit_rate = [1] * ats.num_states

    # example: APs
    ats.add_ap_annotation(
        umbi.ats.AtomicPropositionAnnotation(
            name="is_terminal",
            alias="Terminal State",
            description="Indicates whether the state is terminal.",
        )
    )
    ats.get_ap_annotation("is_terminal").set_state_values(
        [s == 0 or s == ats.num_states - 1 for s in range(ats.num_states)]
    )

    # example: rewards
    ats.add_reward_annotation(
        umbi.ats.RewardAnnotation(
            name="steps",
            alias="step cost",
            description="Cost incurred at each step.",
        )
    )
    ats.get_reward_annotation("steps").set_state_values([-1] * ats.num_states)

    # wall hit penalty
    choice_penalty = [0] * ats.num_choices
    for choice in ats.state_choice_range(0):
        action = ats.choice_to_choice_action[choice]
        if ats.choice_action_to_name[action] == "left":  # left action in first state
            choice_penalty[choice] = -10
    for choice in ats.state_choice_range(ats.num_states - 1):
        action = ats.choice_to_choice_action[choice]
        if ats.choice_action_to_name[action] == "right":  # right action in last state
            choice_penalty[choice] = -10
    ats.add_reward_annotation(
        umbi.ats.RewardAnnotation(
            name="wall_hit_penalty",
            alias="wall hit penalty",
            description="Penalty incurred when hitting a wall.",
        )
    )
    ats.get_reward_annotation("wall_hit_penalty").set_choice_values(choice_penalty)

    # observations: 3 observations, based on state mod 3
    ats.observation_annotation = umbi.ats.ObservationAnnotation(num_observations=3)
    ats.observation_annotation.set_state_values([s % 3 for s in range(ats.num_states)])

    return ats


def main(args):
    ats = random_walk_ats(args.states)
    ats.validate()
    # umbi.io.write_ats(ats, args.output)
    # logger.info(f"Written to {args.output}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Create a random walk umbfile.")
    parser.add_argument("states", help="Number of states", type=int)
    parser.add_argument(
        "--output",
        help="Destination to write to",
        type=pathlib.Path,
        required=False,
        default=pathlib.Path("out.umb"),
    )
    main(parser.parse_args())



from dataclasses import dataclass


@dataclass
class ModelInfo:
    """Container to store information about the model."""

    name: str | None = None
    version: str | None = None
    authors: list[str] | None = None
    description: str | None = None
    comment: str | None = None
    doi: str | None = None
    url: str | None = None



"""
Utilities for reading/writing tarfiles.
"""

import io as std_io
import logging
import pathlib
import tarfile
from typing_extensions import Literal

logger = logging.getLogger(__name__)


class TarReader:
    """
    An auxiliary class to simplify tar reading.
    #TODO lazy loading
    """

    @staticmethod
    def tar_read(tarpath: str | pathlib.Path) -> dict[str, bytes]:
        """
        Load all files from a tarball into memory.
        :return: a dictionary filename -> bytestring
        """
        logger.debug(f"loading tarfile from {tarpath} ...")
        filename_data = {}
        with tarfile.open(tarpath, mode="r:*") as tar:
            for member in tar.getmembers():
                if member.isfile():
                    fileobj = tar.extractfile(member)
                    if fileobj is None:
                        raise KeyError(f"Could not extract file {member.name} from {tarpath}")
                    filename_data[member.name] = fileobj.read()
        logger.debug("tarfile successfully loaded")
        return filename_data

    def __init__(self, tarpath: str | pathlib.Path) -> None:
        self.tarpath = tarpath
        self.filename_bytes = TarReader.tar_read(tarpath)

    @property
    def filenames(self) -> list[str]:
        """List of filenames in the tarfile."""
        return list(self.filename_bytes.keys())

    def has_file(self, filename: str) -> bool:
        """Check if the tarfile contains a file with the given name."""
        return filename in self.filename_bytes

    def read_file(self, filename: str, required: bool = False) -> bytes | None:
        """
        Read raw bytes from a specific file in the tarfile.
        :param filename: name of the file to read
        :param required: whether the file is required
        :return: bytestring, or None if the optional file is not found
        :raises KeyError: if the required file is not found
        """
        if not self.has_file(filename):
            if not required:
                return None
            else:
                raise KeyError(f"tarfile {self.tarpath} has no file {filename}")
        logger.debug(f"loading {filename} ...")
        return self.filename_bytes[filename]


class TarWriter:
    """An auxiliary class to simplify tar writing."""

    @classmethod
    def tar_write(
        cls,
        tarpath: str | pathlib.Path,
        filename_data: dict[str, bytes],
        compression: Literal["gz", "bz2", "xz"] | None = "gz",
    ):
        """
        Create a tarball file with the given contents.

        :param tarpath: path to a tarball file
        :param filename_data: a dictionary filename -> bytestring
        :param compression: compression algorithm or None for no compression
        """
        logger.debug(f"writing tarfile {tarpath} with compression '{compression}' ...")
        mode = "w"
        if compression is not None:
            mode = f"w:{compression}"
        with tarfile.open(tarpath, mode=mode) as tar:  # type: ignore
            for filename, data in filename_data.items():
                tar_info = tarfile.TarInfo(name=filename)
                tar_info.size = len(data)
                tar.addfile(tar_info, std_io.BytesIO(data))
        logger.debug("successfully wrote the tarfile")

    def __init__(self) -> None:
        self.filename_bytes = {}

    @property
    def filenames(self) -> list[str]:
        """List of filenames added to the tarfile."""
        return list(self.filename_bytes.keys())

    def has_file(self, filename: str) -> bool:
        """Check if a file with the given name has been added to the tarfile."""
        return filename in self.filename_bytes

    def add_file(self, filename: str, data: bytes):
        """Add a binary file to the tarfile."""
        logger.debug(f"adding {filename} ...")
        if self.has_file(filename):
            logger.warning(f"file {filename} already exists in the tarfile, overwriting")
        self.filename_bytes[filename] = data

    def write(self, tarpath: str | pathlib.Path):
        """Write all added files to a tarfile."""
        TarWriter.tar_write(tarpath, self.filename_bytes)



"""
Numeric type: union of numeric primitive types and intervals.
"""

from collections.abc import Collection

from .numeric_primitive import (
    NumericPrimitive,
    NumericPrimitiveType,
    numeric_primitive_promotion_type,
    numeric_primitive_type_of,
    promote_numeric_primitive_to,
)
from .interval import Interval, IntervalType, interval_promotion_type, promote_interval_to

""" Alias for all numeric objects. """
NumericType = NumericPrimitiveType | IntervalType

""" Alias for all numeric types. """
Numeric = NumericPrimitive | Interval


def numeric_type_of(value: Numeric) -> NumericType:
    """Determine the numeric type of a given value."""
    if isinstance(value, NumericPrimitive):
        return numeric_primitive_type_of(value)
    else:  # isinstance(value, Interval):
        return value.type


def numeric_promotion_type(types: Collection[NumericType]) -> NumericType:
    """Determine the common numeric type from a set of numeric types. Used for type promotion."""
    assert len(types) > 0, "cannot determine common numeric type of empty set"
    assert all(isinstance(t, (NumericPrimitiveType, IntervalType)) for t in types), (
        f"non-numeric types found in set: {types}"
    )
    if any(isinstance(t, IntervalType) for t in types):
        # convert all primitive types to intervals
        interval_types = list({IntervalType(t) if isinstance(t, NumericPrimitiveType) else t for t in types})
        return interval_promotion_type(interval_types)
    # assert all(isinstance(t, NumericPrimitiveType) for t in types), f"non-primitive types found in set: {types}"
    return numeric_primitive_promotion_type(types)  # type: ignore


def promote_numeric_to(value: Numeric, target_type: NumericType) -> Numeric:
    """Promote a numeric value to the target type."""
    if isinstance(target_type, NumericPrimitiveType):
        assert isinstance(value, NumericPrimitive), f"cannot promote {value} to primitive type {target_type}"
        return promote_numeric_primitive_to(value, target_type)
    else:  # is_interval_type(target_type):
        if not isinstance(value, Interval):
            value = Interval(value, value)
        return promote_interval_to(value, target_type)



"""
Base JSON schema classes and utilities.
"""

import logging
from dataclasses import dataclass
from types import SimpleNamespace
from typing import Any

from marshmallow import (
    INCLUDE,
    Schema,
    ValidationError,
    fields,
    post_load,
)

import umbi.datatypes

logger = logging.getLogger(__name__)


class FieldUint(fields.Int):
    """Custom marshmallow field for unsigned integers."""

    def _deserialize(self, value: Any, attr: Any, data: Any, **kwargs: Any) -> Any:
        result = super()._deserialize(value, attr, data, **kwargs)
        if result is None:
            raise ValidationError("value is required")
        if result < 0:
            raise ValidationError(f"value {value} must be an unsigned integer")
        return result


class JsonSchema(Schema):
    """An abstract class to represent specific schemas that will follow."""

    # to allow unknown fields in the input data
    class Meta:
        unknown = INCLUDE

    @classmethod
    def schema_class(cls) -> type:
        """The class used as a representation of this schema result."""
        return JsonSchemaResult

    @post_load
    def make_object(self, data: Any, **kwargs: Any) -> Any:
        """Create an object with attributes matching all the json fields. Notify about unrecognized fields."""
        extra_fields = set(data.keys()) - set(self.fields.keys())
        for f in extra_fields:
            # TODO issue error in strict mode
            logger.warning(f"JSON contains unrecognized field: {f}")

        obj = self.schema_class()()
        for field in self.fields:
            value = data.get(field, None)
            setattr(obj, field, value)
        return obj

    @classmethod
    def parse(cls, json_data: Any, *args: Any, **kwargs: Any) -> Any:
        """
        Parse from a json object.
        :raises: ValidationError if the json object does not conform to the schema
        """
        try:
            return cls().load(json_data, *args, **kwargs)
        except ValidationError as e:
            logger.error(f"{cls} validation error")
            # messages is actually a json object, so we can pretty print it
            logger.error(umbi.datatypes.json_to_string(e.messages))  # type: ignore
            raise e


@dataclass
class JsonSchemaResult(SimpleNamespace):
    @classmethod
    def class_schema(cls) -> type:
        """The schema responsible for serialization of this class."""
        return JsonSchema

    @classmethod
    def from_json(cls, json_obj):
        return cls.class_schema()().parse(json_obj)

    def to_json(self):
        """Convert the current object to json and strip null values."""
        json_obj = self.class_schema()().dump(self)
        json_obj = umbi.datatypes.json_remove_none_dict_values(json_obj)
        return json_obj

    def __str__(self) -> str:
        """Convert to a string (json format)."""
        return umbi.datatypes.json_to_string(self.to_json())

    def validate(self):
        """Validate the current object syntactically."""
        self.from_json(self.to_json())



"""
Main UMB index class and schema.
"""

from dataclasses import dataclass, field

from marshmallow import fields

from .annotations import AnnotationCategoryToAnnotationMap, AnnotationDescription
from .file_data import FileData, FileDataSchema
from .json_schema import (
    JsonSchema,
    JsonSchemaResult,
    FieldUint,
)
from .model_data import ModelData, ModelDataSchema
from .transition_system import TransitionSystem, TransitionSystemSchema
from .valuations import ValuationDescription, ValuationsSchema


class UmbIndexSchema(JsonSchema):
    """UMB index file schema."""

    format_version = FieldUint(data_key="format-version", required=True)
    format_revision = FieldUint(data_key="format-revision", required=True)
    file_data = fields.Nested(FileDataSchema, data_key="file-data", required=False)
    model_data = fields.Nested(ModelDataSchema, data_key="model-data", required=False)
    transition_system = fields.Nested(TransitionSystemSchema, data_key="transition-system", required=True)
    annotations = AnnotationCategoryToAnnotationMap(data_key="annotations", required=False)
    valuations = ValuationsSchema(data_key="valuations", required=False)

    @classmethod
    def schema_class(cls) -> type:
        return UmbIndex


@dataclass
class UmbIndex(JsonSchemaResult):
    format_version: int = 0
    format_revision: int = 0
    file_data: FileData | None = None
    model_data: ModelData | None = None
    transition_system: TransitionSystem = field(default_factory=TransitionSystem)
    annotations: dict[str, dict[str, AnnotationDescription]] | None = None
    valuations: dict[str, ValuationDescription] | None = None

    @classmethod
    def class_schema(cls) -> type:
        return UmbIndexSchema



"""
Reading and writing umbfiles into ExplicitUmb.
"""

import pathlib
from enum import Enum
import logging

import umbi.binary
import umbi.datatypes
import umbi.umb.index

from umbi.umb import ExplicitUmb
from umbi.binary.sized_type import UINT32, UINT64
from umbi.datatypes import PrimitiveType
from .tar_coders import TarDecoder, TarEncoder


logger = logging.getLogger(__name__)


class UmbFile(Enum):
    """Enum for common files in a umbfile."""

    INDEX = "index.json"
    STATE_IS_INITIAL = "state-is-initial.bin"
    STATE_TO_CHOICES = "state-to-choices.bin"
    STATE_TO_PLAYER = "state-to-player.bin"
    STATE_IS_MARKOVIAN = "state-is-markovian.bin"
    STATE_TO_EXIT_RATE = "state-to-exit-rate.bin"
    CHOICE_TO_BRANCHES = "choice-to-branches.bin"
    BRANCH_TO_TARGET = "branch-to-target.bin"
    BRANCH_TO_PROBABILITY = "branch-to-probability.bin"


class UmbReader(TarDecoder):
    def __init__(self, tarpath: str | pathlib.Path, strict_mode: bool = False) -> None:
        """
        :param tarpath: path to the umbfile
        :param strict_mode: in the strict mode, unknown files will raise an error
        """
        super().__init__(tarpath)
        # to keep track of which files were read
        self.filename_read = {filename: False for filename in self.filenames}
        # in strict mode, we will raise an error if there are any unrecognized files in the umbfile
        self.strict_mode = strict_mode

    def list_unread_files(self):
        """Print warning about unread files from the tarfile, if such exist."""
        unread_files = [f for f, read in self.filename_read.items() if not read]
        for f in unread_files:
            logger.warning(f"umbfile contains unrecognized file: {f}")
        if self.strict_mode and len(unread_files) > 0:
            raise ValueError(f"unrecognized files in umbfile: {unread_files}")

    def read_file(self, filename: str, required: bool = False) -> bytes | None:
        """Read raw bytes from a specific file in the tarball. Mark the file as read."""
        if self.has_file(filename):
            self.filename_read[filename] = True
        return super().read_file(filename, required)

    def read_index_file(self, umb: ExplicitUmb):
        json_bytes = self.read_file(UmbFile.INDEX.value, required=True)
        assert json_bytes is not None
        json_str = umbi.binary.bytes_to_scalar(json_bytes, PrimitiveType.STRING)
        assert isinstance(json_str, str)
        json_obj = umbi.datatypes.string_to_json(json_str)
        pretty_str = umbi.datatypes.json_to_string(json_obj)
        logger.debug(f"loaded the following index file:\n{pretty_str}")
        idx = umbi.umb.index.UmbIndex.from_json(json_obj)
        idx.validate()
        umb.index = idx

    def read_state_files(self, umb: ExplicitUmb):
        """Read state-related files."""
        ts = umb.index.transition_system

        if ts.num_initial_states == 0:
            umb.state_is_initial = [False] * ts.num_states
        else:
            umb.state_is_initial = self.read_bitvector(UmbFile.STATE_IS_INITIAL.value, ts.num_states)

        umb.state_to_choices = self.read_vector(UmbFile.STATE_TO_CHOICES.value, UINT64, required=False)
        if ts.num_players > 0:
            umb.state_to_player = self.read_vector(UmbFile.STATE_TO_PLAYER.value, UINT32, required=False)

        if ts.time == "discrete":
            # ignore the file, all states are probabilistic
            umb.state_is_markovian = [False] * ts.num_states
        else:
            if self.has_file(UmbFile.STATE_IS_MARKOVIAN.value):
                umb.state_is_markovian = self.read_bitvector(UmbFile.STATE_IS_MARKOVIAN.value, ts.num_states)
            else:
                default = ts.time == "stochastic"
                umb.state_is_markovian = [default] * ts.num_states

        if ts.time == "stochastic" and ts.exit_rate_type is not None:
            if self.has_file(UmbFile.STATE_TO_EXIT_RATE.value):
                umb.state_to_exit_rate = self.read_vector(
                    UmbFile.STATE_TO_EXIT_RATE.value, ts.exit_rate_type, required=True
                )

    def read_actions(self, umb: ExplicitUmb, applies_to: str, num_items: int):
        """
        Read action files for a specific action type (choices or branches).
        :param umb: ExplicitUmb object to populate
        :param applies_to: "choices" or "branches"
        :param num_items: number of items (choices or branches)
        """
        path = f"actions/{applies_to}"
        filename = f"{path}/values.bin"

        if not self.has_file(filename):
            # default: everything maps to action 0
            item_to_actions = [0] * num_items
        else:
            item_to_actions = self.read_vector(filename, UINT32)

        if applies_to == "choices":
            umb.choice_to_choice_action = item_to_actions
        else:  # branches
            umb.branch_to_branch_action = item_to_actions

        filename_csr = f"{path}/string-mapping.bin"
        if self.has_file(filename_csr):
            action_to_string = self.read_strings(f"{path}/strings.bin", required=True, filename_csr=filename_csr)
            if applies_to == "choices":
                umb.choice_action_to_string = action_to_string
            else:  # branches
                umb.branch_action_to_string = action_to_string

    def read_transition_files(self, umb: ExplicitUmb):
        """Read transition-related files."""
        umb.choice_to_branches = self.read_vector(UmbFile.CHOICE_TO_BRANCHES.value, UINT64, required=False)
        umb.branch_to_target = self.read_vector(UmbFile.BRANCH_TO_TARGET.value, UINT64, required=False)
        ts = umb.index.transition_system

        if ts.branch_probability_type is not None:
            umb.branch_to_probability = self.read_vector(
                UmbFile.BRANCH_TO_PROBABILITY.value, ts.branch_probability_type, required=True
            )

        if ts.num_choice_actions > 0:
            self.read_actions(umb, "choices", ts.num_choices)

        if ts.num_branch_actions > 0:
            self.read_actions(umb, "branches", ts.num_branches)

    def read_annotation_files(self, umb: ExplicitUmb):
        if umb.index.annotations is None:
            return
        umb.annotations = dict()
        for group, annotation_map in umb.index.annotations.items():
            umb.annotations[group] = dict()
            for name, annotation in annotation_map.items():
                self.read_annotation(umb, group, name, annotation)

    def read_annotation(
        self,
        umb: ExplicitUmb,
        group: str,
        name: str,
        annotation: umbi.umb.index.AnnotationDescription,
    ):
        """
        Read annotation files for a single annotation.
        :param group: annotation group, e.g. "rewards", "aps", or custom group
        :param name: annotation name
        :param annotation: annotation info (Annotation)
        :return: dict mapping applies_to -> values
        """
        ts = umb.index.transition_system
        assert annotation.type is not None, "annotation.type must be specified"
        assert annotation.applies_to is not None and len(annotation.applies_to) > 0, (
            "annotation.applies_to must be a non-empty list"
        )
        assert umb.annotations is not None, "umb.annotations must be specified"
        assert umb.annotations.get(group) is not None, f"missing annotation group {group} in umb.annotations"
        umb.annotations[group][name] = dict()
        if annotation.probability_type is not None:
            raise NotImplementedError("reading stochastic annotations is not implemented")
        for applies_to in annotation.applies_to:
            path = f"annotations/{group}/{name}/{applies_to}"
            if annotation.type.type == PrimitiveType.STRING:
                vector = self.read_strings(
                    f"{path}/strings.bin",
                    required=True,
                    filename_csr=f"{path}/string-mapping.bin",
                )
            elif annotation.type.type == PrimitiveType.BOOL:
                num_entries = {
                    "states": ts.num_states,
                    "choices": ts.num_choices,
                    "branches": ts.num_branches,
                }[applies_to]
                vector = self.read_bitvector(f"{path}/values.bin", num_entries=num_entries)
            else:
                vector = self.read_vector(f"{path}/values.bin", annotation.type, required=True)
            assert isinstance(vector, list), "expected a list"
            umb.annotations[group][name][applies_to] = vector

    def read_valuation_files(self, umb: ExplicitUmb):
        """Read valuation-related files."""
        if umb.index.valuations is None:
            return
        umb.valuations = dict()
        for applies_to, valuation_desc in umb.index.valuations.items():
            self.read_valuation_description(umb, applies_to, valuation_desc)

    def read_valuation_description(
        self,
        umb: ExplicitUmb,
        applies_to: str,
        valuation_desc: umbi.umb.index.ValuationDescription,
    ):
        ts = umb.index.transition_system
        num_entries = {
            "states": ts.num_states,
            "choices": ts.num_choices,
            "branches": ts.num_branches,
            "observations": ts.num_observations,
            "players": ts.num_players,
        }[applies_to]
        # valuation_desc.unique is ignored
        if valuation_desc.num_strings is not None:
            raise NotImplementedError("reading string valuations is not implemented")
        if len(valuation_desc.classes) != 1:
            raise NotImplementedError("reading multiple valuation classes is not implemented")
        struct_type = valuation_desc.classes[0]
        assert not struct_type.contains_strings, "reading valuations with strings is not implemented"
        bytes = self.read_file(f"valuations/{applies_to}/valuations.bin", required=True)
        assert bytes is not None, "expected non-None bytes"
        expected_size = struct_type.size_bytes * num_entries
        assert len(bytes) == expected_size, (
            f"valuation data length does not match expected size: got {len(bytes)}, expected #num_{applies_to}*(struct_size_bytes)={num_entries}*{struct_type.size_bytes} = {expected_size}"
        )
        valuations = umbi.binary.bytes_to_vector(bytes, struct_type)
        assert umb.valuations is not None, "umb.valuations must have been initialized"
        umb.valuations[applies_to] = valuations

    def read_observation_files(self, umb: ExplicitUmb):
        ts = umb.index.transition_system
        if ts.num_observations == 0:
            return
        if ts.observation_probability_type is not None:
            raise NotImplementedError("reading stochastic observations is not implemented")
        applies_to = ts.observations_apply_to
        umb.entity_to_observation = self.read_vector(f"observations/{applies_to}/values.bin", UINT64, required=True)

    def read_umb(self) -> ExplicitUmb:
        logger.info(f"loading umbfile from {self.tarpath} ...")
        filenames_str = "\n".join(self.filenames)
        logger.debug(f"found the following files in the umbfile:\n{filenames_str}")
        umb = ExplicitUmb()
        self.read_index_file(umb)
        self.read_state_files(umb)
        self.read_transition_files(umb)
        self.read_annotation_files(umb)
        self.read_valuation_files(umb)
        self.read_observation_files(umb)
        self.list_unread_files()
        logger.info("umbfile successfully loaded")
        return umb


class UmbWriter(TarEncoder):
    def add_index_file(self, umb: ExplicitUmb):
        """Add the index file."""
        umb.index.validate()
        json_obj = umb.index.to_json()
        json_str = umbi.datatypes.json_to_string(json_obj)
        json_bytes = umbi.binary.scalar_to_bytes(json_str, PrimitiveType.STRING)
        self.add_file(UmbFile.INDEX.value, json_bytes)

    def add_state_files(self, umb: ExplicitUmb):
        """Add state-related files."""
        self.add_bitvector(UmbFile.STATE_IS_INITIAL.value, umb.state_is_initial, required=True)
        self.add_vector(UmbFile.STATE_TO_CHOICES.value, UINT64, umb.state_to_choices)
        if umb.index.transition_system.num_players > 0:
            self.add_vector(UmbFile.STATE_TO_PLAYER.value, UINT32, umb.state_to_player)

        if umb.index.transition_system.time == "urgent-stochastic":
            state_is_markovian = (
                umb.state_is_markovian
                if umb.state_is_markovian is not None
                else [True] * umb.index.transition_system.num_states
            )
            self.add_bitvector(UmbFile.STATE_IS_MARKOVIAN.value, state_is_markovian)
        if umb.index.transition_system.time != "discrete":
            if umb.index.transition_system.exit_rate_type is not None:
                self.add_vector(
                    UmbFile.STATE_TO_EXIT_RATE.value,
                    umb.index.transition_system.exit_rate_type,
                    umb.state_to_exit_rate,
                )

    def add_transition_files(self, umb: ExplicitUmb):
        """Add transition-related files."""
        self.add_vector(UmbFile.CHOICE_TO_BRANCHES.value, UINT64, umb.choice_to_branches)
        self.add_vector(UmbFile.BRANCH_TO_TARGET.value, UINT64, umb.branch_to_target)
        ts = umb.index.transition_system

        if ts.branch_probability_type is not None:
            self.add_vector(
                UmbFile.BRANCH_TO_PROBABILITY.value,
                ts.branch_probability_type,
                umb.branch_to_probability,
            )

        if ts.num_choice_actions > 0:
            self.add_actions(umb, "choices")

        if ts.num_branch_actions > 0:
            self.add_actions(umb, "branches")

    def add_actions(self, umb: ExplicitUmb, applies_to: str):
        """
        Add action files for a specific action type (choices or branches).
        :param umb: ExplicitUmb object to write from
        :param applies_to: "choices" or "branches"
        """
        path = f"actions/{applies_to}"

        if applies_to == "choices":
            item_to_actions = umb.choice_to_choice_action
            action_to_string = umb.choice_action_to_string
        else:  # branches
            item_to_actions = umb.branch_to_branch_action
            action_to_string = umb.branch_action_to_string

        if item_to_actions is not None:
            self.add_vector(f"{path}/values.bin", UINT32, item_to_actions)

        if action_to_string is not None:
            self.add_strings(f"{path}/strings.bin", action_to_string, f"{path}/string-mapping.bin")

    def add_annotation_files(self, umb: ExplicitUmb):
        """Add annotation files."""
        if umb.index.annotations is None:
            return
        assert umb.annotations is not None, "umb.annotations must be specified"
        for group, annotation_map in umb.index.annotations.items():
            assert group in umb.annotations, f"missing annotation group {group} in umb.annotations"
            for name, annotation in annotation_map.items():
                self.add_annotation(umb, group, name, annotation)

    def add_annotation(
        self,
        umb: ExplicitUmb,
        group: str,
        name: str,
        annotation: umbi.umb.index.AnnotationDescription,
    ):
        """
        Add files for a single annotation.
        :param umb: ExplicitUmb object to write from
        :param group: annotation group, e.g. "rewards", "aps"
        :param name: annotation name
        :param annotation: annotation metadata
        """
        assert annotation.type is not None, "annotation.type must be specified"
        assert annotation.applies_to is not None and len(annotation.applies_to) > 0, (
            "annotation.applies_to must be a non-empty list"
        )
        assert umb.annotations is not None, "umb.annotations must be specified"
        assert umb.annotations.get(group) is not None, f"missing annotation group {group} in umb.annotations"
        assert umb.annotations[group].get(name) is not None, f"missing annotation {group}/{name} in umb.annotations"
        for applies_to in annotation.applies_to:
            path = f"annotations/{group}/{name}/{applies_to}"
            values = umb.annotations[group][name][applies_to]
            if annotation.type.type == PrimitiveType.STRING:
                self.add_strings(f"{path}/strings.bin", values, f"{path}/string-mapping.bin", required=True)
            elif annotation.type.type == PrimitiveType.BOOL:
                self.add_bitvector(f"{path}/values.bin", values, required=True)
            else:
                self.add_vector(f"{path}/values.bin", annotation.type, values, required=True)

    def add_valuation_files(self, umb: ExplicitUmb):
        """Add valuation-related files."""
        if umb.index.valuations is None:
            return
        for applies_to, valuation_desc in umb.index.valuations.items():
            self.add_valuation_description(umb, applies_to, valuation_desc)

    def add_valuation_description(
        self,
        umb: ExplicitUmb,
        applies_to: str,
        valuation_desc: umbi.umb.index.ValuationDescription,
    ):
        ts = umb.index.transition_system
        num_entries = {
            "states": ts.num_states,
            "choices": ts.num_choices,
            "branches": ts.num_branches,
            "observations": ts.num_observations,
            "players": ts.num_players,
        }[applies_to]
        # valuation_desc.unique is ignored
        if valuation_desc.num_strings is not None:
            raise NotImplementedError("writing string valuations is not implemented")
        if len(valuation_desc.classes) != 1:
            raise NotImplementedError("writing multiple valuation classes is not implemented")
        struct_type = valuation_desc.classes[0]
        assert not struct_type.contains_strings, "writing valuations with strings is not implemented"
        assert umb.valuations is not None, "umb.valuations must be specified"
        assert applies_to in umb.valuations, f"missing valuation data for {applies_to}"
        valuations = umb.valuations[applies_to]
        bytes = umbi.binary.vector_to_bytes(valuations, struct_type)
        assert len(bytes) == struct_type.size_bytes * num_entries, "valuation data length does not match expected size"
        self.add_file(f"valuations/{applies_to}/valuations.bin", bytes)

    def add_observation_files(self, umb: ExplicitUmb):
        """Add observation-related files."""
        ts = umb.index.transition_system
        if ts.num_observations == 0:
            return
        if ts.observation_probability_type is not None:
            raise NotImplementedError("writing stochastic observations is not implemented")
        applies_to = ts.observations_apply_to
        assert umb.entity_to_observation is not None, "umb.entity_to_observation must be specified"
        self.add_vector(f"observations/{applies_to}/values.bin", UINT64, umb.entity_to_observation, required=True)

    def write_umb(self, umb: ExplicitUmb, umbpath: str | pathlib.Path):
        logger.info(f"writing umbfile to {umbpath} ...")
        self.add_index_file(umb)
        self.add_state_files(umb)
        self.add_transition_files(umb)
        self.add_annotation_files(umb)
        self.add_valuation_files(umb)
        self.add_observation_files(umb)
        self.write(umbpath)
        logger.info("finished writing the umbfile")


def read_umb(umbpath: str | pathlib.Path) -> ExplicitUmb:
    """Read ExplicitUmb from a umbfile."""
    return UmbReader(umbpath).read_umb()


def write_umb(umb: ExplicitUmb, umbpath: str | pathlib.Path, replace_file_data: bool = True):
    """Write ExplicitUmb to a umbfile."""
    if replace_file_data:
        umb.index.file_data = umbi.umb.index.umbi_file_data()
    UmbWriter().write_umb(umb, umbpath)



"""
Interval datatypes.
"""

from dataclasses import dataclass
from fractions import Fraction
from collections.abc import Collection
from .numeric_primitive import (
    NumericPrimitiveType,
    NumericPrimitive,
    numeric_primitive_promotion_type,
    promote_numeric_primitive_to,
)


@dataclass(frozen=True)
class IntervalType:
    base_type: NumericPrimitiveType

    def __str__(self) -> str:
        return f"{self.base_type.value}-interval"

    @classmethod
    def from_string(cls, s: str) -> "IntervalType":
        """Parse an interval type from a string s of type <base-type>-interval."""
        if not s.endswith("-interval"):
            raise ValueError(f"invalid interval type string: {s}")
        base_type_string = s.removesuffix("-interval")
        base_type = NumericPrimitiveType(base_type_string)
        return cls(base_type)


class Interval:
    """Represents a numeric interval where left <= right."""

    def __init__(self, left: NumericPrimitive, right: NumericPrimitive) -> None:
        self.left = left
        self.right = right
        self.validate()

    def validate(self) -> None:
        if not self.left <= self.right:
            raise ValueError(f"expected {self.left} <=  {self.right}")

    def __str__(self) -> str:
        return f"interval[{self.left},{self.right}]"

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, NumericPrimitive):
            other = Interval(other, other)
        if not isinstance(other, Interval):
            return False
        return self.left == other.left and self.right == other.right

    def __contains__(self, value: NumericPrimitive) -> bool:
        """Check if a numeric value is within the interval."""
        return self.left <= value <= self.right

    @property
    def base_type(self) -> NumericPrimitiveType:
        """Get the base numeric type of the interval."""
        if isinstance(self.left, Fraction) or isinstance(self.right, Fraction):
            return NumericPrimitiveType.RATIONAL
        elif isinstance(self.left, float) or isinstance(self.right, float):
            return NumericPrimitiveType.DOUBLE
        else:  # isinstance(self.left, int) and isinstance(self.right, int):
            return NumericPrimitiveType.INT

    @property
    def type(self) -> IntervalType:
        """Get the interval type of the interval."""
        return IntervalType(self.base_type)


def interval_promotion_type(types: Collection[IntervalType]) -> IntervalType:
    """Determine the common interval type from a set of interval types. Used for type promotion."""
    assert len(types) > 0, "cannot determine common numeric type of empty sequence"
    assert all(isinstance(t, IntervalType) for t in types), f"non-interval types found in set: {types}"
    base_types = {t.base_type for t in types}
    common_base_type = numeric_primitive_promotion_type(base_types)
    return IntervalType(common_base_type)


def promote_interval_to(value: Interval, target_type: IntervalType) -> Interval:
    """Promote an interval to the target interval type."""
    target_base_type = target_type.base_type
    return Interval(
        promote_numeric_primitive_to(value.left, target_base_type),
        promote_numeric_primitive_to(value.right, target_base_type),
    )



"""
umbi.binary: Utilities for (de)serializing basic types.

Introduces some composite types.
- SizedType: a scalar type with an associated size (e.g., BOOL-4, UINT-64)
- StructType: a composite type with named attributes of type SizedType
- Struct: a mapping from field names to scalar values, representing an instance of a StructType
"""

from .scalar import bytes_to_scalar, scalar_to_bytes
from .sequence import (
    bytes_to_vector,
    bytes_with_ranges_to_vector,
    vector_to_bytes,
    vector_to_bytes_with_ranges,
)
from .sized_type import (
    SizedType,
    BOOL1,
    UINT32,
    UINT64,
)
from .struct_type import (
    StructType,
    StructPadding,
    StructAttribute,
    Struct,
)

__all__ = [
    # scalar.py
    "scalar_to_bytes",
    "bytes_to_scalar",
    # sequence.py
    "bytes_to_vector",
    "bytes_with_ranges_to_vector",
    "vector_to_bytes",
    "vector_to_bytes_with_ranges",
    # sized_type.py
    "SizedType",
    "BOOL1",
    "UINT32",
    "UINT64",
    # struct_type.py
    "StructType",
    "StructPadding",
    "StructAttribute",
    "Struct",
]



"""
Scalar encompasses primitive and numeric types.
"""

from collections.abc import Collection

from .primitive import Primitive, PrimitiveType, primitive_type_of
from .numeric import NumericType, Numeric, numeric_type_of, numeric_promotion_type, promote_numeric_to

""" Alias for names of scalar types. """
ScalarType = PrimitiveType | NumericType

""" Alias for objects of scalar type. """
Scalar = Primitive | Numeric


def scalar_type_of(value: Scalar) -> ScalarType:
    """Determine the common or interval type of a given value."""
    if isinstance(value, Primitive):
        return primitive_type_of(value)
    else:  # isinstance(value, Numeric):
        return numeric_type_of(value)


def scalar_promotion_type(types: Collection[ScalarType]) -> ScalarType:
    """
    Determine the common type from a set of types. Used for type promotion.
    Rules: bool -> Numeric -> string
    """
    if len(types) == 0:
        raise ValueError("cannot determine common type of empty set")
    if PrimitiveType.STRING in types:
        return PrimitiveType.STRING
    if PrimitiveType.BOOL in types:
        if len(types) == 1:
            return PrimitiveType.BOOL
        types = {t for t in types if t != PrimitiveType.BOOL}
    # assert all(isinstance(t, NumericType) for t in types), f"non-numeric types found in set: {types}"
    return numeric_promotion_type(types)  # type: ignore


def promote_scalar_to(value: Scalar, target_type: ScalarType) -> Scalar:
    """Promote a value to the target type."""
    if target_type == PrimitiveType.STRING:
        return str(value)
    elif target_type == PrimitiveType.BOOL:
        assert isinstance(value, bool), f"cannot promote {value} to bool"
        return value
    else:  # isinstance(target_type, NumericType):
        if isinstance(value, bool):
            value = int(value)
        assert isinstance(value, Numeric), f"cannot promote {value} to numeric type {target_type}"
        return promote_numeric_to(value, target_type)



"""
(De)serialization of sequences of common types or structs.
"""

import logging
from typing import Sequence

from umbi.datatypes import Scalar, ScalarType
from .sized_type import SizedType, BOOL1
from .struct_type import Struct, StructType, bytes_to_struct, struct_to_bytes
from .scalar import (
    bytes_to_scalar,
    scalar_to_bytes,
)

logger = logging.getLogger(__name__)


# Bitvectors


def bytes_to_bitvector(bytestring: bytes) -> list[bool]:
    """Convert a bytestring representing a bitvector into a list of booleans."""
    return [(byte >> bit) & 1 == 1 for byte in bytestring for bit in range(8)]


def bitvector_to_bytes(bitvector: Sequence[bool]) -> bytes:
    """Convert a list of booleans representing a bitvector into a bytestring."""
    byte_array = bytearray()
    for i in range(0, len(bitvector), 8):
        byte = sum((1 << j) for j in range(8) if i + j < len(bitvector) and bitvector[i + j])
        byte_array.append(byte)
    return bytes(byte_array)


# Utils for later


def bytes_with_ranges_into_chunks(data: bytes, chunk_ranges: list[tuple[int, int]]) -> list[bytes]:
    """Split bytestring into chunks according to chunk ranges."""
    assert len(data) == chunk_ranges[-1][1], "data length does not match the end of the last chunk range"
    return [data[start:end] for start, end in chunk_ranges]


def chunks_to_csr(chunks: list[bytes]) -> list[int]:
    """Build csr for a list of chunks."""
    current_pos = 0
    csr = [current_pos]
    for chunk in chunks:
        current_pos += len(chunk)
        csr.append(current_pos)
    return csr


def bytes_into_chunks(data: bytes, chunk_size: int) -> list[bytes]:
    """Split bytestring into evenly sized chunks."""
    assert chunk_size > 0, f"expected {chunk_size} to be a positive number"
    assert len(data) % chunk_size == 0, f"expected {len(data)} to be divisible by {chunk_size}"
    return [data[i : i + chunk_size] for i in range(0, len(data), chunk_size)]


# Vector (de)serialization


def bytes_to_vector(
    data: bytes, value_sized_type: SizedType | StructType, little_endian: bool = True
) -> list[Scalar] | list[Struct]:
    """
    Decode a binary string as a list of elements of the given type.
    :param value_sized_type: vector element (sized) type
    :param little_endian: if True, the binary string is interpreted as little-endian
    """

    if len(data) == 0:
        return []

    if value_sized_type == BOOL1:
        assert little_endian, "big-endianness for bitvectors is not implemented"
        return bytes_to_bitvector(data)  # type: ignore[return-value] (promoting list[bool] to list[Scalar])

    assert value_sized_type.is_byte_aligned, "expected a byte-aligned type as chunk sizes"
    chunk_size = value_sized_type.size_bytes
    chunks = bytes_into_chunks(data, chunk_size)

    if isinstance(value_sized_type, StructType):
        return [bytes_to_struct(chunk, value_sized_type) for chunk in chunks]
    else:
        return [bytes_to_scalar(chunk, value_sized_type.type, little_endian) for chunk in chunks]


def bytes_with_ranges_to_vector(
    data: bytes,
    value_type: ScalarType,
    chunk_ranges: list[tuple[int, int]],
    little_endian: bool = True,
) -> list[Scalar]:
    """
    Decode a binary string as a list of numbers.
    :param value_type: vector element type
    :param chunk_ranges: (optional) chunk ranges to split the data into
    :param little_endian: if True, the binary string is interpreted as little-endian
    """
    if len(data) == 0:
        return []
    chunks = bytes_with_ranges_into_chunks(data, chunk_ranges)
    return [bytes_to_scalar(chunk, value_type, little_endian) for chunk in chunks]


def vector_to_bytes(
    vector: Sequence[Scalar] | Sequence[Struct], value_sized_type: SizedType | StructType, little_endian: bool = True
) -> bytes:
    """Encode a list of values as a binary string.
    :param value_sized_type: vector element type
    :return: encoded binary string
    """

    if len(vector) == 0:
        logger.warning("converting empty vector to bytes")
        return b""

    if value_sized_type == BOOL1:
        assert little_endian, "big-endianness for bitvectors is not implemented"
        assert all(isinstance(v, bool) for v in vector), "expected a list of booleans for BOOL1 type"
        return bitvector_to_bytes(vector)  # type: ignore[arg-type]

    if isinstance(value_sized_type, StructType):
        assert all(isinstance(v, dict) for v in vector), f"expected a list of structs, got {vector}"
        chunks = [struct_to_bytes(item, value_sized_type) for item in vector]  # type: ignore[list-item]
    else:
        assert all(isinstance(v, Scalar) for v in vector), f"expected a list of scalars, got {vector}"
        chunks = [
            scalar_to_bytes(item, value_sized_type.type, value_sized_type.size_bytes, little_endian)  # type: ignore
            for item in vector
        ]
    bytestring = b"".join(chunks)
    expected_size = len(vector) * (value_sized_type.size_bytes)
    assert len(bytestring) == expected_size, f"expected {expected_size} bytes, but got {len(bytestring)} bytes"
    return bytestring


def vector_to_bytes_with_ranges(
    vector: Sequence[Scalar], value_sized_type: SizedType, little_endian: bool = True
) -> tuple[bytes, list[int] | None]:
    """Encode a list of values as a binary string.
    :param value_sized_type: vector element type
    :return: encoded binary string
    :return: (optional) chunk csr if non-trivial splitting is needed to split the resulting bytestring into chunks, e.g. for strings or non-standard rationals
    """

    if len(vector) == 0:
        logger.warning("converting empty vector to bytes")
        return (b"", None)

    chunks = [
        scalar_to_bytes(item, value_sized_type.type, value_sized_type.size_bytes, little_endian) for item in vector
    ]
    chunks_csr = chunks_to_csr(chunks)
    bytestring = b"".join(chunks)

    return bytestring, chunks_csr



"""
Model data schemas and classes.
"""

from dataclasses import dataclass

from marshmallow import fields

from .json_schema import (
    JsonSchema,
    JsonSchemaResult,
)


class ModelDataSchema(JsonSchema):
    name = fields.String(data_key="name", required=False)
    version = fields.String(data_key="version", required=False)
    authors = fields.List(fields.String(), data_key="authors", required=False)
    description = fields.String(data_key="description", required=False)
    comment = fields.String(data_key="comment", required=False)
    doi = fields.String(data_key="doi", required=False)
    url = fields.String(data_key="url", required=False)

    @classmethod
    def schema_class(cls) -> type:
        return ModelData


@dataclass
class ModelData(JsonSchemaResult):
    name: str | None = None
    version: str | None = None
    authors: list[str] | None = None
    description: str | None = None
    comment: str | None = None
    doi: str | None = None
    url: str | None = None

    @classmethod
    def class_schema(cls) -> type:
        return ModelDataSchema



"""Type definitions for sized scalar types."""

from umbi.datatypes import (
    Primitive,
    PrimitiveType,
    NumericPrimitive,
    NumericPrimitiveType,
    Interval,
    IntervalType,
    Scalar,
    ScalarType,
)

from collections.abc import Sequence
from fractions import Fraction
from dataclasses import dataclass

# Sized type definitions


@dataclass
class SizedType:
    """
    Represents a scalar type with an associated size in bits.
    """

    type: ScalarType
    size_bits: int  # size in bits

    def __init__(self, type: ScalarType, size_bits: int | None = None) -> None:
        self.type = type
        if size_bits is None:
            size_bits = scalar_type_default_bitsize(type)
        self.size_bits = size_bits

    def __post_init__(self):
        self.validate()

    def validate(self):
        assert self.size_bits > 0, "size must be positive"
        validate_scalar_type_size(self.type, self.size_bits)

    @property
    def is_byte_aligned(self) -> bool:
        return self.size_bits % 8 == 0

    @property
    def size_bytes(self) -> int:
        return (self.size_bits + 7) // 8  # ceiling division


# Common sized types used in umbi

BOOL1 = SizedType(PrimitiveType.BOOL, 1)
UINT32 = SizedType(NumericPrimitiveType.UINT, 32)
UINT64 = SizedType(NumericPrimitiveType.UINT, 64)


# Default bitsizes for types


def primitive_type_default_bitsize(type: PrimitiveType) -> int:
    """Get the default size in bits for a primitive type."""
    return {
        PrimitiveType.BOOL: 1,
        PrimitiveType.STRING: 64,
    }[type]


def numeric_primitive_type_default_bitsize(type: NumericPrimitiveType) -> int:
    """Get the default size in bits for a numeric primitive type."""
    return {
        NumericPrimitiveType.INT: 64,
        NumericPrimitiveType.UINT: 64,
        NumericPrimitiveType.DOUBLE: 64,
        NumericPrimitiveType.RATIONAL: 128,
    }[type]


def interval_type_default_bitsize(type: IntervalType) -> int:
    """Get the default size in bits for an interval type."""
    return 2 * numeric_primitive_type_default_bitsize(type.base_type)


def scalar_type_default_bitsize(type: ScalarType) -> int:
    """Get the default size in bits for a scalar type."""
    if isinstance(type, PrimitiveType):
        return primitive_type_default_bitsize(type)
    elif isinstance(type, NumericPrimitiveType):
        return numeric_primitive_type_default_bitsize(type)
    else:  # isinstance(type, IntervalType)
        return interval_type_default_bitsize(type)


def primitive_smallest_bitsize(type: PrimitiveType) -> int:
    """Get the default size in bits for a given primitive type."""
    return primitive_type_default_bitsize(type)


# Scalar type size validation


def validate_primitive_type_size(type: PrimitiveType, size: int) -> None:
    """Validate that a size is valid for a primitive type."""
    if type == PrimitiveType.STRING:
        if size != 64:
            raise ValueError(f"{type.value} size must be 64")


def validate_numeric_primitive_type_size(type: NumericPrimitiveType, size: int) -> None:
    """Validate that a size is valid for a numeric primitive type."""
    if type == NumericPrimitiveType.DOUBLE:
        if size != 64:
            raise ValueError(f"{type.value} size must be 64")
    if type == NumericPrimitiveType.RATIONAL:
        if size % 2 != 0:
            raise ValueError(f"{type.value} size must be a multiple of 2")


def validate_interval_type_size(type: IntervalType, size: int) -> None:
    """Validate that a size is valid for an interval type."""
    if size % 2 != 0:
        raise ValueError(f"{str(type)} size must be a multiple of 2")
    if type.base_type == NumericPrimitiveType.RATIONAL and size % 4 != 0:
        raise ValueError(f"{str(type)} size must be a multiple of 4")


def validate_scalar_type_size(type: ScalarType, size: int) -> None:
    """Validate that a size is valid for a scalar type."""
    if isinstance(type, PrimitiveType):
        validate_primitive_type_size(type, size)
    elif isinstance(type, NumericPrimitiveType):
        validate_numeric_primitive_type_size(type, size)
    else:  # isinstance(type, IntervalType)
        validate_interval_type_size(type, size)


# Size computation for values


def num_bits_for_primitive(value: Primitive, type: PrimitiveType) -> int:
    """Calculate the number of bits needed to represent a primitive value."""
    return primitive_type_default_bitsize(type)


def num_bits_for_integer(value: int, signed: bool = True, round_to_8: bool = True) -> int:
    """
    Return the number of bits needed to represent an integer value.
    :param signed: if True, the integer is signed
    :param round_to_8: if True, the number of bits is rounded up to the nearest multiple of 8
    """
    if not signed:
        num_bits = value.bit_length()
    else:
        if value >= 0:
            num_bits = value.bit_length()
        else:
            num_bits = (-value - 1).bit_length()
        num_bits += 1  # add sign bit
    if round_to_8 and num_bits % 8 != 0:
        num_bits += 8 - (num_bits % 8)
    return num_bits


def num_bits_for_numeric_primitive(value: NumericPrimitive, value_type: NumericPrimitiveType) -> int:
    """Calculate the number of bits needed to represent a numeric primitive value."""
    if value_type == NumericPrimitiveType.INT or value_type == NumericPrimitiveType.UINT:
        assert isinstance(value, int), f"expected an int value, got {type(value)}"
        return num_bits_for_integer(value, signed=(value_type == NumericPrimitiveType.INT), round_to_8=False)
    elif value_type == NumericPrimitiveType.DOUBLE:
        return numeric_primitive_type_default_bitsize(value_type)
    else:  # value_type == NumericPrimitiveType.RATIONAL
        assert isinstance(value, Fraction), f"expected a Fraction, got {type(value)}"
        numerator_bits = num_bits_for_integer(value.numerator, signed=True, round_to_8=True)
        denominator_bits = num_bits_for_integer(value.denominator, signed=False, round_to_8=True)
        return max(numerator_bits, denominator_bits) * 2


def num_bits_for_interval(value: Interval, value_type: IntervalType) -> int:
    """Calculate the number of bits needed to represent an interval value."""
    left_bits = num_bits_for_numeric_primitive(value.left, value_type.base_type)
    right_bits = num_bits_for_numeric_primitive(value.right, value_type.base_type)
    return max(left_bits, right_bits) * 2


def num_bits_for_scalar(value: Scalar, value_type: ScalarType) -> int:
    """Calculate the number of bits needed to represent a scalar value."""
    if isinstance(value_type, PrimitiveType):
        assert isinstance(value, Primitive), f"expected a primitive value for type {value_type}, got {type(value)}"
        return num_bits_for_primitive(value, value_type)
    elif isinstance(value_type, NumericPrimitiveType):
        assert isinstance(value, NumericPrimitive), (
            f"expected a numeric primitive value for type {value_type}, got {type(value)}"
        )
        return num_bits_for_numeric_primitive(value, value_type)
    else:  # isinstance(value_type, IntervalType):
        assert isinstance(value, Interval), f"expected an interval value for type {value_type}, got {type(value)}"
        return num_bits_for_interval(value, value_type)


# Size computation for sequences


def max_num_bits_for_sequence_element(values: Sequence[Scalar], value_type: ScalarType) -> int:
    """Calculate the number of bits needed to represent an element of a sequence."""
    assert len(values) > 0, "cannot compute max bitsize for an empty sequence"
    return max(num_bits_for_scalar(v, value_type) for v in values)


# def num_bytes_for_integer(value: int, signed: bool = True, round_to_8: bool = True) -> int:
#     """
#     Return the number of bytes needed to represent an integer value.
#     :param round_to_8: if True, the number of bytes is rounded up to the nearest multiple of 8
#     """
#     num_bytes = num_bits_for_integer(value, signed=signed, round_to_8=round_to_8) // 8  # round up to full bytes
#     if round_to_8 and num_bytes % 8 != 0:
#         num_bytes += 8 - (num_bytes % 8)
#     return num_bytes


# def num_bytes_for_rational(value: Fraction) -> int:
#     """Calculate the number of bytes needed to represent a rational number."""
#     # rounding up to a number of bytes of multiple of 8
#     numerator_size = num_bytes_for_integer(value.numerator, signed=True, round_to_8=True)
#     denominator_size = num_bytes_for_integer(value.denominator, signed=False, round_to_8=True)
#     return max(numerator_size, denominator_size) * 2



"""
Annotation schemas and classes.
"""

from dataclasses import dataclass
from typing import Any, Literal, Type

from marshmallow import fields, validate

from umbi.binary import SizedType

from .type import SizedTypeSchema

from .json_schema import (
    JsonSchema,
    JsonSchemaResult,
)


class AnnotationSchema(JsonSchema):
    """Schema for one annotation."""

    alias = fields.String(data_key="alias", required=False)
    description = fields.String(data_key="description", required=False)
    applies_to = fields.List(
        fields.String(validate=validate.OneOf(["states", "choices", "branches"])),
        data_key="applies-to",
        required=False,
        validate=validate.Length(min=1),
    )
    type = fields.Nested(SizedTypeSchema, data_key="type", required=True)
    lower = fields.Integer(data_key="lower", required=False, validate=validate.Range(-(2**63), 2**63 - 1))
    upper = fields.Integer(data_key="upper", required=False, validate=validate.Range(-(2**63), 2**63 - 1))
    num_strings = fields.Integer(data_key="#strings", required=False, validate=validate.Range(1, 2**31 - 1))
    probability_type = fields.Nested(SizedTypeSchema, data_key="probability-type", required=False)

    @classmethod
    def schema_class(cls) -> Type:
        return AnnotationDescription


@dataclass
class AnnotationDescription(JsonSchemaResult):
    """Annotation data class."""

    alias: str | None = None
    description: str | None = None
    applies_to: list[Literal["states", "choices", "branches"]] | None = None
    type: SizedType | None = None
    lower: float | None = None
    upper: float | None = None
    num_strings: int | None = None
    probability_type: SizedType | None = None

    @classmethod
    def class_schema(cls) -> Type:
        return AnnotationSchema


class AnnotationMap(fields.Dict):
    """Marshmallow field for a map from names to AnnotationDescription values."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(keys=fields.String(), values=fields.Nested(AnnotationSchema), *args, **kwargs)


class AnnotationCategoryToAnnotationMap(fields.Dict):
    """Marshmallow field for a map from annotation category to a name-to-annotation map."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(keys=fields.String(), values=AnnotationMap(), *args, **kwargs)



"""(De)serializers for numeric primitive types."""

import struct
from fractions import Fraction

from bitstring import BitArray

from umbi.datatypes import NumericPrimitive, NumericPrimitiveType

from .utils import split_bits_half, split_bytes_half


# Double


def bytes_to_double(data: bytes, little_endian: bool = True) -> float:
    """Convert a bytestring to a double."""
    ef = "<" if little_endian else ">"
    return struct.unpack(f"{ef}d", data)[0]


def double_to_bytes(value: float, little_endian: bool = True) -> bytes:
    """Convert a double to a bytestring."""
    ef = "<" if little_endian else ">"
    return struct.pack(f"{ef}d", value)


def bits_to_double(bits: BitArray) -> float:
    """Convert bitstring to a double."""
    assert len(bits) == 64, "double must be represented with 64 bits"
    return bits.float


def double_to_bits(value: float) -> BitArray:
    """Convert a double to a bitstring."""
    return BitArray(float=value, length=64)


# Integers


def bytes_to_integer(data: bytes, signed: bool = True, little_endian: bool = True) -> int:
    """Convert a bytestring to an integer."""
    return int.from_bytes(data, byteorder="little" if little_endian else "big", signed=signed)


def integer_to_bytes(value: int, num_bytes: int, signed: bool = True, little_endian: bool = True) -> bytes:
    """
    Convert an integer to a bytestring.
    :raises OverflowError if the integer cannot be represented in the given number of bytes.
    """
    return value.to_bytes(num_bytes, byteorder="little" if little_endian else "big", signed=signed)


def bits_to_integer(bits: BitArray, signed: bool = True) -> int:
    """Convert a bitstring to an integer."""
    if signed:
        return bits.int
    else:
        return bits.uint


def integer_to_bits(value: int, num_bits: int, signed: bool = True) -> BitArray:
    """
    Convert an integer to a bitstring.
    :raises ValueError if the integer cannot be represented in the given number of bits.
    """
    if signed:
        return BitArray(int=value, length=num_bits)
    else:
        return BitArray(uint=value, length=num_bits)


# Rationals
# Convention: a normalized rational has a non-negative denominator.
# Rationals are represented as two integers of equal lengths: a signed numerator and an unsigned denominator.


def normalize_rational(value: Fraction) -> Fraction:
    """Ensure that the denominator of a fraction is non-negative."""
    if value.denominator < 0:
        value = Fraction(-value.numerator, -value.denominator)
    return value


def bytes_to_rational(data: bytes, little_endian: bool = True) -> Fraction:
    """Convert a bytestring to a fraction."""
    numerator_bytes, denominator_bytes = split_bytes_half(data)
    numerator = bytes_to_integer(numerator_bytes, signed=True, little_endian=little_endian)
    denominator = bytes_to_integer(denominator_bytes, signed=False, little_endian=little_endian)
    return Fraction(numerator, denominator)


def rational_to_bytes(value: Fraction, num_bytes: int, little_endian: bool = True) -> bytes:
    """Convert a fraction to a bytestring."""
    value = normalize_rational(value)
    assert num_bytes % 2 == 0, "num_bytes must be even to encode both numerator and denominator"
    term_size_bytes = num_bytes // 2
    # assuming term_size_bytes is large enough to encode the numerator and denominator
    numerator_bytes = integer_to_bytes(
        value.numerator, num_bytes=term_size_bytes, signed=True, little_endian=little_endian
    )
    denominator_bytes = integer_to_bytes(
        value.denominator,
        num_bytes=term_size_bytes,
        signed=False,
        little_endian=little_endian,
    )
    return numerator_bytes + denominator_bytes


def bits_to_rational(bits: BitArray) -> Fraction:
    """Convert a bitstring to a fraction."""
    numerator_bits, denominator_bits = split_bits_half(bits)
    numerator = bits_to_integer(numerator_bits, signed=True)
    denominator = bits_to_integer(denominator_bits, signed=False)
    return Fraction(numerator, denominator)


def rational_to_bits(value: Fraction, num_bits: int) -> BitArray:
    """Convert a fraction to a bitstring."""
    value = normalize_rational(value)
    assert num_bits % 2 == 0, "num_bits must be even to encode both numerator and denominator"
    term_num_bits = num_bits // 2
    numerator_bits = integer_to_bits(value.numerator, term_num_bits, signed=True)
    denominator_bits = integer_to_bits(value.denominator, term_num_bits, signed=False)
    return numerator_bits + denominator_bits


# API


def bytes_to_numeric_primitive(data: bytes, value_type: NumericPrimitiveType, little_endian: bool = True):
    """Convert a bytestring to a numeric primitive value."""
    return {
        NumericPrimitiveType.INT: lambda d: bytes_to_integer(d, signed=True, little_endian=little_endian),
        NumericPrimitiveType.UINT: lambda d: bytes_to_integer(d, signed=False, little_endian=little_endian),
        NumericPrimitiveType.DOUBLE: lambda d: bytes_to_double(d, little_endian=little_endian),
        NumericPrimitiveType.RATIONAL: lambda d: bytes_to_rational(d, little_endian=little_endian),
    }[value_type](data)


def numeric_primitive_to_bytes(
    value: NumericPrimitive, value_type: NumericPrimitiveType, num_bytes: int, little_endian: bool = True
) -> bytes:
    """Convert a numeric primitive value to a bytestring."""
    return {
        NumericPrimitiveType.INT: lambda v: integer_to_bytes(v, num_bytes, signed=True, little_endian=little_endian),
        NumericPrimitiveType.UINT: lambda v: integer_to_bytes(v, num_bytes, signed=False, little_endian=little_endian),
        NumericPrimitiveType.DOUBLE: lambda v: double_to_bytes(v, little_endian=little_endian),
        NumericPrimitiveType.RATIONAL: lambda v: rational_to_bytes(v, num_bytes, little_endian=little_endian),
    }[value_type](value)


def bits_to_numeric_primitive(bits: BitArray, value_type: NumericPrimitiveType) -> NumericPrimitive:
    """Convert a bitstring to a numeric primitive value."""
    return {
        NumericPrimitiveType.INT: lambda b: bits_to_integer(b, signed=True),
        NumericPrimitiveType.UINT: lambda b: bits_to_integer(b, signed=False),
        NumericPrimitiveType.DOUBLE: bits_to_double,
        NumericPrimitiveType.RATIONAL: bits_to_rational,
    }[value_type](bits)


def numeric_primitive_to_bits(value: NumericPrimitive, value_type: NumericPrimitiveType, num_bits: int) -> BitArray:
    """Convert a numeric primitive value to a bitstring."""
    return {
        NumericPrimitiveType.INT: lambda v: integer_to_bits(v, num_bits, signed=True),
        NumericPrimitiveType.UINT: lambda v: integer_to_bits(v, num_bits, signed=False),
        NumericPrimitiveType.DOUBLE: double_to_bits,
        NumericPrimitiveType.RATIONAL: lambda v: rational_to_bits(v, num_bits),
    }[value_type](value)


__all__ = [
    "bytes_to_numeric_primitive",
    "numeric_primitive_to_bytes",
    "bits_to_numeric_primitive",
    "numeric_primitive_to_bits",
]



"""
Auxiliary collection operations.
"""

from .scalar import ScalarType, Scalar, scalar_promotion_type, scalar_type_of, promote_scalar_to

from collections.abc import Sequence, Collection


def collection_types(collection: Collection[Scalar]) -> set[ScalarType]:
    """Determine the set of scalar types of elements in the collection."""
    return {scalar_type_of(x) for x in collection}


def collection_promotion_type(collection: Collection[Scalar]) -> ScalarType:
    """Determine the common type to which all elements in the collection can be promoted."""
    return scalar_promotion_type(collection_types(collection))


def promote_vector_to(vector: Sequence[Scalar], target_type: ScalarType) -> list[Scalar]:
    """Promote a vector of scalars to the target type."""
    return [promote_scalar_to(v, target_type) for v in vector]


def promote_vector(vector: Sequence[Scalar]) -> tuple[ScalarType, list[Scalar]]:
    """
    Promote a vector of scalars to the common promotion type.
    :returns: the promotion type and the promoted vector
    """
    target_type = collection_promotion_type(vector)
    vector = promote_vector_to(vector, target_type)
    return target_type, vector



"""
umbi.datatypes: Core datatype definitions and utilities.

This package contains the fundamental datatype classes and definitions used throughout umbi,
including intervals, composite types, vector utilities, and JSON operations.

Scalar types: the following describe enums for names of scalar types and their python equivalents:
- PrimitiveType and Primitive: primitive scalar types (e.g., BOOL, STRING)
- NumericPrimitiveType and NumericPrimitive: numeric primitive types (e.g., INT, FLOAT, RATIONAL)
- IntervalType and Interval: interval types (e.g., INT_INTERVAL)
- NumericType and Numeric: numeric primitive and interval types
- ScalarType and Scalar: all scalar types (primitive, numeric primitive, and interval)

TODO: consider moving json-related utilities outside of datatypes
"""

from .primitive import PrimitiveType, Primitive
from .numeric_primitive import NumericPrimitiveType, NumericPrimitive
from .interval import IntervalType, Interval
from .numeric import NumericType, Numeric
from .scalar import ScalarType, Scalar, scalar_promotion_type
from .collection import collection_promotion_type, promote_vector_to, promote_vector

# json.py
from .json import (
    JsonLike,
    json_remove_none_dict_values,
    json_to_string,
    string_to_json,
)


__all__ = [
    # primitive.py
    "PrimitiveType",
    "Primitive",
    # numeric_primitive.py
    "NumericPrimitiveType",
    "NumericPrimitive",
    # interval.py
    "IntervalType",
    "Interval",
    # numeric.py
    "NumericType",
    "Numeric",
    # scalar.py
    "ScalarType",
    "Scalar",
    "scalar_promotion_type",
    # vector.py
    "collection_promotion_type",
    "promote_vector_to",
    "promote_vector",
    # json.py
    "JsonLike",
    "json_remove_none_dict_values",
    "json_to_string",
    "string_to_json",
]



"""
umbi.io: Utilities for (de)serializating umfiles and ATSs.
"""

from .umb import read_umb, write_umb
from .umb_ats_converter import explicit_umb_to_explicit_ats, explicit_ats_to_explicit_umb, read_ats, write_ats

__all__ = [
    "read_umb",
    "write_umb",
    "explicit_umb_to_explicit_ats",
    "explicit_ats_to_explicit_umb",
    "read_ats",
    "write_ats",
]



"""Explicit in-memory representation of a umbfile."""

from dataclasses import dataclass, field

from umbi.datatypes import Numeric

from .index.umb_index import UmbIndex


@dataclass
class ExplicitUmb:
    """Explicit representation of a umbfile using Python collections."""

    index: UmbIndex = field(default_factory=UmbIndex)

    state_is_initial: list[bool] = field(default_factory=list)
    state_to_choices: list[int] | None = None
    state_to_player: list[int] | None = None

    state_is_markovian: list[bool] | None = None
    state_to_exit_rate: list[Numeric] | None = None

    choice_to_branches: list[int] | None = None
    branch_to_target: list[int] | None = None
    branch_to_probability: list[Numeric] | None = None

    choice_to_choice_action: list[int] | None = None
    choice_action_to_string: list[str] | None = None

    branch_to_branch_action: list[int] | None = None
    branch_action_to_string: list[str] | None = None

    annotations: dict[str, dict[str, dict[str, list]]] | None = None  # group -> name -> applies_to -> values
    valuations: dict[str, list] | None = None  # applies_to -> valuations
    entity_to_observation: list[int] | None = (
        None  # for each entity (index.observations_apply_to), an observation index
    )

    def validate(self):
        self.index.validate()



"""
Transition system schemas and classes.
"""

from dataclasses import dataclass
from typing import Literal, Type

from marshmallow import fields, validate

from .type import SizedTypeSchema

from umbi.binary import SizedType

from .json_schema import (
    JsonSchema,
    JsonSchemaResult,
)


class TransitionSystemSchema(JsonSchema):
    """Transition system schema."""

    time = fields.String(
        data_key="time",
        required=True,
        validate=validate.OneOf(["discrete", "stochastic", "urgent-stochastic"]),
    )
    num_players = fields.Int(data_key="#players", required=True, validate=validate.Range(min=0))
    num_states = fields.Int(data_key="#states", required=True, validate=validate.Range(min=1))
    num_initial_states = fields.Int(data_key="#initial-states", required=True, validate=validate.Range(min=0))
    num_choices = fields.Int(data_key="#choices", required=True, validate=validate.Range(min=0))
    num_choice_actions = fields.Int(data_key="#choice-actions", required=True, validate=validate.Range(min=0))
    num_branches = fields.Int(data_key="#branches", required=True, validate=validate.Range(min=0))
    num_branch_actions = fields.Int(data_key="#branch-actions", required=True, validate=validate.Range(min=0))
    num_observations = fields.Int(data_key="#observations", required=True, validate=validate.Range(min=0))
    observations_apply_to = fields.String(
        data_key="observations-apply-to",
        required=False,
        validate=validate.OneOf(["states", "branches"]),  # why can't observation apply to choices?
    )

    branch_probability_type = fields.Nested(SizedTypeSchema, data_key="branch-probability-type", required=False)
    exit_rate_type = fields.Nested(SizedTypeSchema, data_key="exit-rate-type", required=False)
    observation_probability_type = fields.Nested(
        SizedTypeSchema, data_key="observation-probability-type", required=False
    )
    player_to_name = fields.List(fields.String(), data_key="player-names", required=False)

    @classmethod
    def schema_class(cls) -> type:
        return TransitionSystem


@dataclass
class TransitionSystem(JsonSchemaResult):
    """Transition system data class."""

    time: Literal["discrete", "stochastic", "urgent-stochastic"] = "discrete"
    num_players: int = 0
    num_states: int = 1
    num_initial_states: int = 0
    num_choices: int = 0
    num_choice_actions: int = 0
    num_branches: int = 0
    num_branch_actions: int = 0
    num_observations: int = 0
    observations_apply_to: Literal["states", "branches"] | None = None

    branch_probability_type: SizedType | None = None
    exit_rate_type: SizedType | None = None
    observation_probability_type: SizedType | None = None
    player_to_name: list[str] | None = None

    @classmethod
    def class_schema(cls) -> Type:
        return TransitionSystemSchema



from dataclasses import dataclass, field


from umbi.datatypes import ScalarType, Scalar, NumericPrimitiveType, collection_promotion_type

from collections.abc import Iterable

from .entity_class import EntityClass

from .domain import Domain


@dataclass
class Variable:
    """Variable data class."""

    # the name of the variable
    _name: str
    # sorted list of possible values
    _domain: Domain | None = None

    def __init__(self, name: str) -> None:
        self._name = name

    def __hash__(self) -> int:
        return hash(self._name)

    @property
    def name(self) -> str:
        return self._name

    @property
    def has_domain(self) -> bool:
        return self._domain is not None

    def _assert_domain_set(self) -> None:
        if self._domain is None:
            raise ValueError(f"Variable '{self.name}' has no domain defined.")

    def invalidate_domain(self) -> None:
        self._domain = None

    def set_domain(self, domain: Domain) -> None:
        self._domain = domain
        self._domain.sort()

    def set_values(self, values: Iterable[Scalar]) -> None:
        self.set_domain(Domain(values))

    @property
    def domain(self) -> Domain:
        self._assert_domain_set()
        assert self._domain is not None
        return self._domain

    @property
    def promotion_type(self) -> ScalarType:
        self._assert_domain_set()
        assert self._domain is not None
        return collection_promotion_type(self._domain)

    @property
    def lower(self) -> Scalar:
        self._assert_domain_set()
        assert self._domain is not None
        return self._domain.lower

    @property
    def upper(self) -> Scalar:
        self._assert_domain_set()
        assert self._domain is not None
        return self._domain.upper

    def __str__(self) -> str:
        if not self.has_domain:
            return f"Variable(name={self.name}, type=?, domain=?)"
        if self.type == NumericPrimitiveType.INT and len(self.domain) == (self.upper - self.lower + 1):  # type: ignore
            domain_str = f"[{self.lower}..{self.upper}]"
        else:
            domain_str = str(self.domain)
        return f"Variable(name={self.name}, type={self.promotion_type}, domain={domain_str})"

    def __repr__(self) -> str:
        return self.__str__()

    def validate(self) -> None:
        pass


@dataclass
class VariableValuations:
    """Mapping from entity to a given variable valuations."""

    # the variable
    _variable: Variable
    # for each entity, the valuation of the variable
    _values: list[Scalar | None] = field(default_factory=lambda: [])

    @property
    def variable(self) -> Variable:
        return self._variable

    @property
    def values(self) -> list[Scalar | None]:
        return self._values

    @property
    def num_values(self) -> int:
        return len(self._values)

    def __str__(self) -> str:
        return f"VariableValuations(variable={self.variable}, values={self.values})"

    def __repr__(self) -> str:
        return self.__str__()

    def ensure_capacity(self, num_entities: int) -> None:
        """Ensures that the valuations list has at least num_entities entries."""
        while len(self._values) < num_entities:
            self._values.append(None)

    def get_entity_value(self, entity: int) -> Scalar | None:
        """Gets the valuation for a given entity index."""
        if entity < 0 or entity >= self.num_values:
            raise IndexError(f"entity index {entity} out of range [0, {self.num_values})")
        return self._values[entity]

    def set_entity_value(self, entity: int, value: Scalar | None) -> None:
        """Sets the valuation for a given entity index. Increases capacity if needed."""
        self.ensure_capacity(entity + 1)
        self._values[entity] = value

    @property
    def has_undefined_values(self) -> bool:
        """Checks if there are any undefined (None) values."""
        return any(v is None for v in self._values)

    def sync_domain(self) -> None:
        """Sets the variable domain from the valuations."""
        if self.has_undefined_values:
            raise ValueError(f"The domain cannot be synced: entity {self._values.index(None)} has undefined value.")
        self._variable.set_values(self._values)  # type: ignore

    def validate(self) -> None:
        self.sync_domain()
        self._variable.validate()


# alias for a mapping from variable to valuation for a single entity
EntityValuation = dict[Variable, Scalar | None]


@dataclass
class EntityValuations:
    """Maintains a collection of VariableValuations."""

    # number of entities (states, actions, etc.) to be associated with variable valuations
    _num_entities: int = 0
    # for each variable name, the corresponding Variable
    _variable_name_to_variable: dict[str, Variable] = field(default_factory=lambda: {})
    # for each variable, the corresponding VariableValuations
    _variable_to_valuations: dict[Variable, VariableValuations] = field(default_factory=lambda: {})

    @property
    def variables(self) -> list[Variable]:
        """Returns a list of all variables defined."""
        return list(self._variable_to_valuations.keys())

    def __str__(self) -> str:
        lines = [f"{self.__class__.__name__}(num_entities={self.num_entities}):"]
        for variable in self.variables:
            variable_valuation = self.get_variable_valuations(variable)
            lines.append(f"  {variable_valuation}")
        return "\n".join(lines)

    @property
    def num_variables(self) -> int:
        """Returns the number of variables defined."""
        return len(self.variables)

    @property
    def num_entities(self) -> int:
        return self._num_entities

    def assert_entity_index_in_range(self, entity: int) -> None:
        if entity < 0 or entity >= self.num_entities:
            raise IndexError(f"entity {entity} out of range [0, {self.num_entities})")

    def has_variable(self, variable_name: str) -> bool:
        """Checks if a VariableValuation exists for the given variable name."""
        return variable_name in self._variable_name_to_variable

    def add_variable(self, variable_name: str) -> Variable:
        """Adds a new VariableValuation for a given variable."""
        if self.has_variable(variable_name):
            raise ValueError(f"Variable '{variable_name}' already exists.")
        variable = Variable(name=variable_name)
        self._variable_name_to_variable[variable_name] = variable
        self._variable_to_valuations[variable] = VariableValuations(variable)
        return variable

    def get_variable(self, variable_name: str) -> Variable:
        """Retrieves the Variable for a given variable name."""
        if not self.has_variable(variable_name):
            raise KeyError(f"Variable '{variable_name}' not found.")
        return self._variable_name_to_variable[variable_name]

    def get_or_add_variable(self, variable_name: str) -> Variable:
        """Retrieves the Variable for a given variable name, adding it if it does not exist."""
        if not self.has_variable(variable_name):
            return self.add_variable(variable_name)
        return self.get_variable(variable_name)

    def remove_variable(self, variable: Variable) -> None:
        """Removes the VariableValuations for a given variable."""
        if not self.has_variable(variable.name):
            raise KeyError(f"Variable '{variable.name}' not found.")
        variable = self._variable_name_to_variable[variable.name]
        del self._variable_name_to_variable[variable.name]
        del self._variable_to_valuations[variable]

    def get_variable_valuations(self, variable: Variable) -> VariableValuations:
        """Retrieves the VariableValuations for a given variable."""
        if not self.has_variable(variable.name):
            raise KeyError(f"Variable '{variable.name}' not found.")
        variable = self._variable_name_to_variable[variable.name]
        return self._variable_to_valuations[variable]

    def ensure_capacity(self, num_entities: int) -> None:
        """Ensures that all VariableValuations have at least num_entities entries."""
        for variable_valuation in self._variable_to_valuations.values():
            variable_valuation.ensure_capacity(num_entities)
        if self._num_entities < num_entities:
            self._num_entities = num_entities

    def get_entity_valuation(self, entity: int) -> EntityValuation:
        """Gets the variable valuations for a given entity index."""
        return {
            variable: variable_valuation.get_entity_value(entity)
            for variable, variable_valuation in self._variable_to_valuations.items()
        }

    def set_entity_valuation(self, entity: int, valuation: EntityValuation) -> None:
        """Adds a new entity with the given variable valuations."""
        self.ensure_capacity(entity + 1)
        for variable, value in valuation.items():
            variable_valuation = self.get_variable_valuations(variable)
            variable_valuation.set_entity_value(entity, value)

    def remove_entity(self, entity: int) -> None:
        """Removes the valuations for a given entity index."""
        raise NotImplementedError

    @property
    def has_distinct_valuations(self) -> bool:
        """Determines whether the variable valuations are distinct across all entities."""
        valuations = [self.get_entity_valuation(entity) for entity in range(self.num_entities)]
        valuations = [tuple(valuation.items()) for valuation in valuations]
        return len(valuations) == len(set(valuations))

    def sync_domains(self) -> None:
        """Synchronizes the domains of all variables based on their valuations."""
        for variable_valuation in self._variable_to_valuations.values():
            variable_valuation.sync_domain()

    def validate(self) -> None:
        self.sync_domains()
        for variable_valuation in self._variable_to_valuations.values():
            if not variable_valuation.num_values == self.num_entities:
                raise ValueError(
                    f"Variable '{variable_valuation.variable.name}' has {variable_valuation.num_values} "
                    f"entities, expected {self.num_entities}"
                )
            variable_valuation.validate()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, EntityValuations):
            return False
        if self.num_entities != other.num_entities:
            return False
        # TODO more thorough comparison
        return True


class EntityClassValuations(dict[EntityClass, EntityValuations]):
    """Mapping from entity class to variable valuations."""

    @property
    def has_values(self) -> bool:
        """Check if the annotation has any values set."""
        return len(self) > 0

    @property
    def entity_classes(self) -> set[EntityClass]:
        """Get the set of entity classes for which this annotation has values."""
        return set(self.keys())

    def has_values_for(self, entity_class: EntityClass) -> bool:
        """Check if the annotation has values for the given entity class."""
        return self.get(entity_class) is not None

    def get_valuations_for(self, entity_class: EntityClass) -> EntityValuations:
        """
        Get the values for the given entity class.
        :raises KeyError: if no values are set for the given entity class
        """
        if not self.has_values_for(entity_class):
            raise KeyError(f"Annotation has no values for entity class {entity_class}")
        return self[entity_class]

    def set_valuations_for(self, entity_class: EntityClass, values: EntityValuations) -> None:
        """Set the values for the given entity class."""
        self[entity_class] = values

    def unset_valuations_for(self, entity_class: EntityClass) -> None:
        """Unset the values for the given entity class."""
        if not self.has_values_for(entity_class):
            raise KeyError(f"Annotation has no values for entity class {entity_class}")
        del self[entity_class]

    def __str__(self) -> str:
        lines = [f"{self.__class__.__name__}:"]
        for entity_class, valuations in self.items():
            lines.append(f"  {entity_class}: {valuations}")
        return "\n".join(lines)

    ### Convenience properties and methods for each entity class ###
    @property
    def has_state_valuations(self) -> bool:
        return self.has_values_for(EntityClass.STATES)

    @property
    def has_choice_valuations(self) -> bool:
        return self.has_values_for(EntityClass.CHOICES)

    @property
    def has_branch_valuations(self) -> bool:
        return self.has_values_for(EntityClass.BRANCHES)

    @property
    def has_observation_valuations(self) -> bool:
        return self.has_values_for(EntityClass.OBSERVATIONS)

    @property
    def has_player_valuations(self) -> bool:
        return self.has_values_for(EntityClass.PLAYERS)

    @property
    def state_valuations(self) -> EntityValuations:
        return self.get_valuations_for(EntityClass.STATES)

    @property
    def choice_valuations(self) -> EntityValuations:
        return self.get_valuations_for(EntityClass.CHOICES)

    @property
    def branch_valuations(self) -> EntityValuations:
        return self.get_valuations_for(EntityClass.BRANCHES)

    @property
    def observation_valuations(self) -> EntityValuations:
        return self.get_valuations_for(EntityClass.OBSERVATIONS)

    @property
    def player_valuations(self) -> EntityValuations:
        return self.get_valuations_for(EntityClass.PLAYERS)

    def set_state_valuations(self, values: EntityValuations):
        self.set_valuations_for(EntityClass.STATES, values)

    def set_choice_valuations(self, values: EntityValuations):
        self.set_valuations_for(EntityClass.CHOICES, values)

    def set_branch_valuations(self, values: EntityValuations):
        self.set_valuations_for(EntityClass.BRANCHES, values)

    def set_observation_valuations(self, values: EntityValuations):
        self.set_valuations_for(EntityClass.OBSERVATIONS, values)

    def set_player_valuations(self, values: EntityValuations):
        self.set_valuations_for(EntityClass.PLAYERS, values)

    def unset_state_valuations(self):
        self.unset_valuations_for(EntityClass.STATES)

    def unset_choice_valuations(self):
        self.unset_valuations_for(EntityClass.CHOICES)

    def unset_branch_valuations(self):
        self.unset_valuations_for(EntityClass.BRANCHES)

    def unset_observation_valuations(self):
        self.unset_valuations_for(EntityClass.OBSERVATIONS)

    def unset_player_valuations(self):
        self.unset_valuations_for(EntityClass.PLAYERS)



from importlib import import_module

from .logger import set_log_level, setup_logging
from .version import __format_revision__, __format_version__, __toolname__, __version__

_SUBMODULES = {"ats", "binary", "datatypes", "io", "umb", "version"}


def __getattr__(name: str):
    if name in _SUBMODULES:
        module = import_module(f".{name}", __name__)
        globals()[name] = module
        return module
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    return sorted(list(globals().keys()) + list(_SUBMODULES))


__all__ = [
    "set_log_level",
    "setup_logging",
    "__toolname__",
    "__version__",
    "__format_version__",
    "__format_revision__",
    "version",
]



#!/usr/bin/env python3
"""
umbi demo: A random stochastic game generator based on Azeem et al. "Optimistic and Topological Value Iteration for
Simple Stochastic Games".
"""

import argparse
import logging
import pathlib
import random
import time

import umbi

logger = logging.getLogger(__name__)


def fill_action(s: int, action: int, num_states: int, delta: dict) -> list[tuple[int, float]]:
    """
    Create a valid probability distribution for an action from state s.

    :param s: source state.
    :param action: action identifier.
    :param num_states: total number of states.
    :param delta: transition function delta[s][a] = [(target, prob), ...].
    :return: list of (target_state, probability) tuples.
    """
    # initialize state entry if needed
    if s not in delta:
        delta[s] = {}
    if action not in delta[s]:
        delta[s][action] = []

    branches = delta[s][action]
    existing_targets = {target for target, _ in branches}

    # maintain a list of available targets and update it incrementally
    available_targets = [sp for sp in range(num_states) if sp not in existing_targets]

    while available_targets:
        # pick a random state s' from available targets
        idx = random.randrange(len(available_targets))
        sp = available_targets[idx]

        # increase delta[s][a] with sp by a random number in (0, 1]
        prob = random.uniform(0.01, 1.0)
        branches.append((sp, prob))

        # remove the chosen target from available list
        available_targets[idx] = available_targets[-1]
        available_targets.pop()

        # calculate total probability
        total_prob = sum(p for _, p in branches)

        # check termination conditions
        if total_prob > 1:
            # decrease the most recently modified probability so sum = 1
            branches[-1] = (sp, prob - (total_prob - 1))
            break
        elif total_prob == 1 or not available_targets:
            # either sum is 1 or all states have been assigned
            if total_prob < 1:
                # increase the most recently modified probability so sum = 1
                branches[-1] = (sp, prob + (1 - total_prob))
            break

    return branches


def random_transition_function(
    num_states: int, seed: int | None = None, max_actions_per_state: int = 3
) -> tuple[dict, dict[int, list[int]], list[str]]:
    """
    Generate a random transition function for a stochastic game.

    :param num_states: number of states (n).
    :param seed: random seed for reproducibility.
    :param max_actions_per_state: maximum number of actions per state.
        - delta: transition function delta[s][a] = [(target, prob), ...]
        - actions_at_state: mapping from state to list of available actions
        - all_actions: list of action names
    """
    if seed is not None:
        random.seed(seed)

    # enumerate states in a random order
    state_order = list(range(num_states))
    random.shuffle(state_order)

    # track transitions: delta[s][a] = [(target, prob), ...]
    delta = {}
    # track actions available at each state: Av(s)
    actions_at_state = {s: [] for s in range(num_states)}
    # track which states have incoming transitions for fast lookup
    states_with_incoming = [False] * num_states
    # global action list
    all_actions = []
    action_counter = 0

    # forward procedure: ensure connectivity from initial state
    logger.info("starting forward procedure ...")
    forward_start = time.time()
    for i in range(1, num_states):
        s = state_order[i]

        # if s has an incoming transition then skip
        if states_with_incoming[s]:
            continue

        # pick any state s' with index smaller than s
        sp = random.choice(state_order[:i])

        # create an action a that starts at s'
        action = action_counter
        action_counter += 1
        all_actions.append(f"fwd_a{action}")

        # assign to (s', a) a positive probability of reaching s and create valid distribution
        branches = fill_action(sp, action, num_states, delta)

        # ensure s is reachable with positive probability
        if not any(target == s for target, _ in branches):
            # force s to be reachable
            if len(branches) == 0:
                branches.append((s, 1.0))
            else:
                # redistribute to make room for s
                scale = 0.5
                branches = [(t, p * scale) for t, p in branches]
                branches.append((s, 1.0 - scale))
            delta[sp][action] = branches

        # mark all target states as having incoming transitions
        for target, _ in branches:
            states_with_incoming[target] = True

        # add action to Av(s')
        actions_at_state[sp].append(action)

    forward_time = time.time() - forward_start
    logger.info(f"forward procedure completed in {forward_time:.3f}s")

    # backward procedure: add as many actions as possible
    logger.info("starting backward procedure ...")
    backward_start = time.time()
    for i in range(num_states - 1, -1, -1):
        s = state_order[i]

        # pick a random number m to determine how many actions to add
        m = random.randint(1, max_actions_per_state)

        # every state must have at least one action
        if len(actions_at_state[s]) == 0:
            m = max(m, 1)

        # add m actions
        for _ in range(m):
            action = action_counter
            action_counter += 1
            all_actions.append(f"bwd_a{action}")

            # fill the action with valid probability distribution
            branches = fill_action(s, action, num_states, delta)

            # mark all target states as having incoming transitions
            for target, _ in branches:
                states_with_incoming[target] = True

            # add action to Av(s)
            actions_at_state[s].append(action)

    backward_time = time.time() - backward_start
    logger.info(f"backward procedure completed in {backward_time:.3f}s")

    return delta, actions_at_state, all_actions


def random_game_ats(num_states: int, seed: int | None = None) -> umbi.ats.ExplicitAts:
    """
    Generate a random stochastic game connected from initial state.

    :param num_states: Number of states (n).
    :param seed: Random seed for reproducibility.
    :return: ExplicitAts representing the random stochastic game.
    """
    delta, actions_at_state, all_actions = random_transition_function(num_states, seed)

    # create ATS
    ats = umbi.ats.ExplicitAts(
        time=umbi.ats.TimeType.STOCHASTIC,
        num_states=num_states,
        num_players=2,
    )

    # partition S uniformly at random into player states and probabilistic states
    # player states are controlled by player 0, probabilistic states by player 1
    is_player0_state = [random.choice([True, False]) for _ in range(num_states)]

    # set player information
    ats.state_to_player = [0 if is_player0 else 1 for is_player0 in is_player0_state]

    # set initial state to the state with index 0
    ats.set_initial_states([0])

    # build the ATS structure from delta and actions_at_state
    ats.num_choice_actions = len(all_actions)
    ats.choice_action_to_name = all_actions

    # count choices and branches
    ats.num_choices = sum(len(actions) for actions in actions_at_state.values())
    ats.num_branches = sum(
        len(delta.get(s, {}).get(action, [])) for s in range(num_states) for action in actions_at_state[s]
    )

    # build CSR structures
    logger.info("building CSR structures ...")
    build_start = time.time()
    ats.state_to_choice = []
    ats.choice_to_choice_action = []
    ats.choice_to_branches = []
    ats.branch_to_target = []
    ats.branch_to_probability = []

    for s in range(num_states):
        ats.state_to_choice.append(len(ats.choice_to_choice_action))

        for action in sorted(actions_at_state[s]):
            ats.choice_to_choice_action.append(action)
            ats.choice_to_branches.append(len(ats.branch_to_target))

            # get all branches for this (s, action) pair
            branches = delta[s][action]
            branches.sort()  # sort by target state

            for target, prob in branches:
                ats.branch_to_target.append(target)
                ats.branch_to_probability.append(prob)

    ats.state_to_choice.append(len(ats.choice_to_choice_action))
    ats.choice_to_branches.append(len(ats.branch_to_target))

    build_time = time.time() - build_start
    logger.info(f"CSR building completed in {build_time:.3f}s")
    logger.info(f"generated random game: {num_states} states, {ats.num_choices} choices, {ats.num_branches} branches")
    logger.info(f"player0 states: {sum(is_player0_state)}, player1 states: {num_states - sum(is_player0_state)}")

    return ats


def main(args):
    ats = random_game_ats(args.states, args.seed)
    ats.validate()
    # umbi.io.write_ats(ats, args.output)
    # logger.info(f"Written to {args.output}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Create a random stochastic game umbfile.")
    parser.add_argument("states", help="Number of states", type=int)
    parser.add_argument(
        "--output",
        help="Destination to write to",
        type=pathlib.Path,
        required=False,
        default=pathlib.Path("out.umb"),
    )
    parser.add_argument(
        "--seed",
        help="Random seed for reproducibility",
        type=int,
        default=None,
    )
    main(parser.parse_args())



import logging
from umbi.datatypes import Scalar

logger = logging.getLogger(__name__)

# from collections.abc import Collection


class Domain(set[Scalar]):
    """
    A set of values with additional information about the range. The collection can be freely modified, calling sort()
    caches the sorted domain.
    """

    # a sorted list of possible values
    _sorted_domain: list[Scalar] | None = None

    def invalidate(self):
        """Invalidate the cached domain information."""
        self._sorted_domain = None

    @property
    def sorted(self) -> bool:
        return self._sorted_domain is not None

    @property
    def sorted_domain(self) -> list[Scalar]:
        if self._sorted_domain is None:
            raise ValueError("the domain is not sorted")
        return self._sorted_domain

    @property
    def can_be_sorted(self) -> bool:
        try:
            sorted(self)  # type: ignore
            return True
        except TypeError:
            return False

    def sort(self):
        """Sorts the domain."""
        self._sorted_domain = None
        try:
            self._sorted_domain = sorted(self)  # type: ignore
        except TypeError:
            logger.warning("domain values cannot be sorted")

    @property
    def lower(self) -> Scalar:
        if self._sorted_domain is None:
            raise ValueError("the domain is not sorted")
        return self._sorted_domain[0]

    @property
    def upper(self) -> Scalar:
        if self._sorted_domain is None:
            raise ValueError("the domain is not sorted")
        return self._sorted_domain[-1]



"""Utility functions for byte-level operations."""

from bitstring import BitArray


def split_bytes(bytestring: bytes, length: int) -> tuple[bytes, bytes]:
    """Split a bytestring into chunks of the given size."""
    assert length > 0, "chunk size must be positive"
    assert len(bytestring) >= length, "data is shorter than the specified length"
    return bytestring[:length], bytestring[length:]


def split_bytes_half(bytestring: bytes) -> tuple[bytes, bytes]:
    """Split a bytestring into two equal halves."""
    assert len(bytestring) % 2 == 0, "bytestring length must be even to split in half"
    mid = len(bytestring) // 2
    return bytestring[:mid], bytestring[mid:]


def split_bits(bits: BitArray, length: int) -> tuple[BitArray, BitArray]:
    """Split a bitstring into chunks of the given size."""
    assert length > 0, "chunk size must be positive"
    assert len(bits) >= length, "data is shorter than the specified length"
    return bits[:length], bits[length:]


def split_bits_half(bits: BitArray) -> tuple[BitArray, BitArray]:
    """Split a bitstring into two equal halves."""
    assert len(bits) % 2 == 0, "bitstring length must be even to split in half"
    mid = len(bits) // 2
    return bits[:mid], bits[mid:]



"""
Utility functions for vector handling (CSR and ranges).
"""


def is_vector_ranges(ranges: list[tuple[int, int]]) -> bool:
    """Check if a vector is a list of ranges."""
    if len(ranges) == 0:
        return False

    def is_interval(x: tuple[int, int]) -> bool:
        return x[0] <= x[1]

    if not all(is_interval(interval) for interval in ranges):
        return False
    for i in range(len(ranges) - 1):
        if ranges[i][1] != ranges[i + 1][0]:
            return False
    return True


def is_vector_csr(vector: list[int]) -> bool:
    """Check if a vector is a CSR row start index vector."""
    if len(vector) < 2:
        return False
    if vector[0] != 0:
        return False
    if not all(vector[i] <= vector[i + 1] for i in range(len(vector) - 1)):
        return False
    return True


def csr_to_ranges(csr: list[int]) -> list[tuple[int, int]]:
    """Convert row start indices to ranges."""
    assert is_vector_csr(csr), "input is not a valid CSR vector"
    ranges = [(csr[i], csr[i + 1]) for i in range(len(csr) - 1)]
    return ranges


def ranges_to_csr(ranges: list[tuple[int, int]]) -> list[int]:
    """Convert ranges to CSR row start indices."""
    assert is_vector_ranges(ranges), "input is not a valid ranges vector"
    csr = [interval[0] for interval in ranges]
    csr.append(ranges[-1][-1])
    return csr



"""(De)serialization for scalar types."""

from umbi.datatypes import (
    Primitive,
    PrimitiveType,
    NumericPrimitive,
    NumericPrimitiveType,
    Interval,
    IntervalType,
    Scalar,
    ScalarType,
)

from bitstring import BitArray

from .primitive import (
    bytes_to_primitive,
    primitive_to_bytes,
    bits_to_primitive,
    primitive_to_bits,
)
from .numeric_primitive import (
    bytes_to_numeric_primitive,
    numeric_primitive_to_bytes,
    bits_to_numeric_primitive,
    numeric_primitive_to_bits,
)
from .utils import split_bytes_half


# Intervals


def bytes_to_interval(data: bytes, value_type: IntervalType, little_endian: bool = True) -> Interval:
    """Convert a bytestring to an interval."""
    lower, upper = split_bytes_half(data)
    base_type = value_type.base_type
    lower = bytes_to_numeric_primitive(lower, base_type, little_endian)
    upper = bytes_to_numeric_primitive(upper, base_type, little_endian)
    return Interval(lower, upper)


def interval_to_bytes(
    interval: Interval, value_type: IntervalType, num_bytes: int, little_endian: bool = True
) -> bytes:
    """Convert an interval to a bytestring."""
    assert num_bytes % 2 == 0, "expected an even-sized num_bytes for interval"
    term_num_bytes = num_bytes // 2
    base_type = value_type.base_type
    lower = numeric_primitive_to_bytes(interval.left, base_type, term_num_bytes, little_endian)
    upper = numeric_primitive_to_bytes(interval.right, base_type, term_num_bytes, little_endian)
    return lower + upper


def bits_to_interval(bits: BitArray, value_type: IntervalType) -> Interval:
    """Convert a bitstring to an interval."""
    raise NotImplementedError("unpacking intervals from bitstrings is not supported in umbi")


def interval_to_bits(interval: Interval, value_type: IntervalType, num_bits: int) -> BitArray:
    """Convert an interval to a bitstring."""
    raise NotImplementedError("packing intervals into bitstrings is not supported in umbi")


# API


def bytes_to_scalar(data: bytes, value_type: ScalarType, little_endian: bool = True) -> Scalar:
    """Convert a binary string to a single value of the given common type."""
    if isinstance(value_type, PrimitiveType):
        return bytes_to_primitive(data, value_type)
    else:  # umbi.datatypes.is_numeric_type(value_type):
        if isinstance(value_type, NumericPrimitiveType):
            return bytes_to_numeric_primitive(data, value_type, little_endian=little_endian)
        else:  # umbi.datatypes.is_interval_type(value_type):
            assert isinstance(value_type, IntervalType), "expected an interval type"
            return bytes_to_interval(data, value_type, little_endian=little_endian)


def scalar_to_bytes(
    value: Scalar, value_type: ScalarType, num_bytes: int | None = None, little_endian: bool = True
) -> bytes:
    """
    Convert a scalar to a bytestring.
    :note: this method intentionally uses SizedType instead of ScalarType for ease of use outside this subpackage
    """
    # expected_type = {
    #     PrimitiveType: Primitive,
    #     NumericPrimitiveType: NumericPrimitive,
    #     IntervalType: Interval,
    # }[value_type.__class__]
    # assert isinstance(value, expected_type), f"expected a value of type {expected_type} for type {value_type}, got {type(value)}"
    if isinstance(value_type, PrimitiveType):
        # assert num_bytes is None, "num_bytes should not be provided for primitive types"
        assert isinstance(value, Primitive), "expected a primitive value"
        return primitive_to_bytes(value, value_type)
    else:  # umbi.datatypes.is_numeric_type(value_type):
        assert num_bytes is not None, "num_bytes must be provided for numeric types"
        if isinstance(value_type, NumericPrimitiveType):
            assert isinstance(value, NumericPrimitive), "expected a numeric value"
            return numeric_primitive_to_bytes(value, value_type, num_bytes, little_endian=little_endian)
        else:  # umbi.datatypes.is_interval_type(value_type):
            assert isinstance(value_type, IntervalType), "expected an interval type"
            assert isinstance(value, Interval), "expected an interval value"
            return interval_to_bytes(value, value_type, num_bytes, little_endian=little_endian)


def bits_to_scalar(bits: BitArray, value_type: ScalarType) -> Scalar:
    """Convert a BitArray to a single value of the given common type."""
    if isinstance(value_type, PrimitiveType):
        return bits_to_primitive(bits, value_type)
    else:  # umbi.datatypes.is_numeric_type(value_type):
        if isinstance(value_type, NumericPrimitiveType):
            return bits_to_numeric_primitive(bits, value_type)
        else:  # umbi.datatypes.is_interval_type(value_type):
            return bits_to_interval(bits, value_type)


def scalar_to_bits(value: Scalar, value_type: ScalarType, num_bits: int) -> BitArray:
    """Convert a scalar to a bitstring."""
    if isinstance(value_type, PrimitiveType):
        assert isinstance(value, Primitive), "expected a primitive value"
        return primitive_to_bits(value, value_type, num_bits)
    else:  # umbi.datatypes.is_numeric_type(value_type):
        if isinstance(value_type, NumericPrimitiveType):
            assert isinstance(value, NumericPrimitive), "expected a numeric value"
            return numeric_primitive_to_bits(value, value_type, num_bits)
        else:  # umbi.datatypes.is_interval_type(value_type):
            assert isinstance(value, Interval), "expected an interval value"
            # assert isinstance(value_type, IntervalType), "expected an interval type"
            return interval_to_bits(value, value_type, num_bits)


__all__ = [
    "bytes_to_scalar",
    "scalar_to_bytes",
    "bits_to_scalar",
    "scalar_to_bits",
]



#!/usr/bin/env python3
"""
umbi demo: Create an ATS from a grid string.
"""

import argparse
import pathlib
import sys
from fractions import Fraction

import umbi
import logging

logger = logging.getLogger(__name__)


def grid_ats_from_string(grid: str) -> umbi.ats.ExplicitAts:
    """
    Create a simple ATS from a rectangular grid string.

    :param grid: A rectangular grid string where:
        - 'i' represents the initial state
        - 'g' represents a goal state
        - 'x' represents an obstacle
        - any other character represents an empty cell
    :return: An ExplicitAts object.
    """
    # parse the grid into a 2D array
    parsed_grid = [
        list(row.strip()) for row in reversed(grid.strip().split("\n"))
    ]  # reverse to have (0,0) at bottom left
    rows, cols = len(parsed_grid), len(parsed_grid[0])
    if any(len(row) != cols for row in parsed_grid):
        raise ValueError("The grid must be rectangular (all rows must have the same length).")

    # Create ATS
    ats = umbi.ats.ExplicitAts()
    ats.time = umbi.ats.TimeType.DISCRETE
    ats.num_players = 1

    cell_to_state = {}
    initial_states = set()
    goal_states = set()

    state_to_cell = []

    for y in range(rows):
        for x in range(cols):
            c = parsed_grid[y][x]
            if c == "x":
                continue
            state = len(cell_to_state)
            state_to_cell.append((x, y))
            cell_to_state[(x, y)] = state
            if c == "i":
                initial_states.add(state)
            elif c == "g":
                goal_states.add(state)

    ats.num_states = len(cell_to_state)
    ats.set_initial_states(list(initial_states))

    if ats.variable_valuations is None:
        ats.variable_valuations = umbi.ats.EntityClassValuations()
    if not ats.variable_valuations.has_state_valuations:
        ats.variable_valuations.set_state_valuations(umbi.ats.EntityValuations())
    state_valuations = ats.variable_valuations.state_valuations
    vx = state_valuations.add_variable("x")
    vy = state_valuations.add_variable("y")
    for state in range(ats.num_states):
        x, y = state_to_cell[state]
        state_valuations.set_entity_valuation(state, {vx: x, vy: y})

    direction_dxdy = {
        "up": (0, 1),
        "down": (0, -1),
        "left": (-1, 0),
        "right": (1, 0),
    }

    ats.state_to_choice = []
    ats.choice_to_branches = []
    ats.branch_to_target = []
    ats.branch_to_probability = []
    ats.choice_to_choice_action = []
    ats.choice_action_to_name = list(direction_dxdy.keys())
    ats.num_choice_actions = len(ats.choice_action_to_name)

    for (x, y), state in cell_to_state.items():
        ats.state_to_choice.append(len(ats.choice_to_choice_action))

        for direction, (dx, dy) in direction_dxdy.items():
            target = (x + dx, y + dy)
            ats.choice_to_choice_action.append(ats.choice_action_to_name.index(direction))
            ats.choice_to_branches.append(len(ats.branch_to_target))

            if target in cell_to_state:
                target_state = cell_to_state[target]
                ats.branch_to_target.extend([target_state, state])
                ats.branch_to_probability.extend([Fraction(9, 10), Fraction(1, 10)])
            else:
                ats.branch_to_target.append(state)
                ats.branch_to_probability.append(1)

    ats.state_to_choice.append(len(ats.choice_to_choice_action))
    ats.choice_to_branches.append(len(ats.branch_to_target))
    ats.num_choices = len(ats.choice_to_choice_action)
    ats.num_branches = len(ats.branch_to_target)

    goal_ap = umbi.ats.AtomicPropositionAnnotation(name="goal")
    goal_ap.set_state_values([s in goal_states for s in range(ats.num_states)])
    ats.add_ap_annotation(goal_ap)

    step_cost_reward = umbi.ats.RewardAnnotation(name="step_cost")
    step_cost_reward.set_choice_values([1 for _ in range(ats.num_choices)])
    ats.add_reward_annotation(step_cost_reward)

    return ats


def main(args):
    if args.input == "-":
        text = sys.stdin.read()
    else:
        with pathlib.Path(args.input).open("rt") as f:
            text = f.read()
    ats = grid_ats_from_string(text)
    ats.validate()
    # umbi.io.write_ats(ats, args.output)
    # logger.info(f"Written to {args.output}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Create an umbfile from a gridworld text file.")
    parser.add_argument(
        "input", help="filename of the gridworld text file (or '-' to read from stdin)", type=str, default="-"
    )
    parser.add_argument(
        "--output",
        help="Destination to write to",
        type=pathlib.Path,
        required=False,
        default=pathlib.Path("out.umb"),
    )
    main(parser.parse_args())



"""
Primitive scalar datatypes.
#TODO: introduce StringType to include strings and jsons?
"""

import enum

""" Primitive scalar type names. """


class PrimitiveType(str, enum.Enum):
    BOOL = "bool"
    STRING = "string"


""" Alias for primitive scalar objects. """
Primitive = bool | str


def primitive_type_of(value: Primitive) -> PrimitiveType:
    """Determine the primitive type of a given value."""
    if isinstance(value, bool):
        return PrimitiveType.BOOL
    else:  # isinstance(value, str):
        return PrimitiveType.STRING
